import sys
import numpy as np
from numba import njit
import env

game_name = sys.argv[1]
env.make(game_name)


from numba.typed import Dict

CHAIN_LENGTH = 4


def DataAgent():
    perData = Dict()
    temp_ = Dict()
    temp_[0] = np.full(0, 0, dtype=np.int64)
    temp_.clear()

    perData[0] = temp_.copy()  #  Các biến dùng trong thuật toán
    perData[1] = temp_.copy()  #  Temp bias của các chuỗi
    perData[2] = temp_.copy()  #  Per bias của các chuỗi
    perData[3] = temp_.copy()  #  Đếm số lần xuất hiện của các chuỗi

    #  Các biến dùng trong thuật toán
    perData[0][0] = np.full(1, 0, dtype=np.int64)  #  Đếm số trận
    perData[0][1] = np.full(CHAIN_LENGTH, 0, dtype=np.int64)  #  Lưu lịch sử
    perData[0][2] = np.full(1, 0, dtype=np.int64)  #  Đếm số turn trong trận
    perData[0][3] = np.full(10000, -1, dtype=np.int64)  #  Lưu các chuỗi xuất hiện

    #  Bias tự do (Temp)
    perData[1][-1] = np.arange(env.getActionSize(), dtype=np.int64) + 1
    np.random.shuffle(perData[1][-1])

    #  Bias tự do (Per)
    perData[2][-1] = np.full(env.getActionSize(), 0, dtype=np.int64) + 1

    #
    return perData


@njit()
def encode(arr_chain):
    res = 0
    for i in range(arr_chain.shape[0]):
        res += arr_chain[i] * env.getActionSize() ** i

    return res


@njit()
def Train(state, perData):
    count_match = perData[0][0]
    action_history = perData[0][1]
    count_turn = perData[0][2]
    chain_history = perData[0][3]

    temp_bias = perData[1]
    per_bias = perData[2]
    count_bias = perData[3]

    reward = env.getReward(state)
    validActions = env.getValidActions(state)

    if count_match[0] < 10000:
        if count_match[0] == -1:
            #  print("ahihi")
            arr_key = np.full(len(count_bias), 0, dtype=np.int64)
            arr_value = np.full(len(count_bias), 0, dtype=np.int64)
            i = 0
            for key in count_bias:
                arr_key[i] = key
                arr_value[i] = count_bias[key][0]
                i += 1

            for i in range(arr_key.shape[0]):
                if arr_value[i] <= 1:
                    #  print(arr_key[i], arr_value[i], "drop")
                    count_bias.pop(arr_key[i])
                    temp_bias.pop(arr_key[i])
                    per_bias.pop(arr_key[i])
                else:
                    count_bias[arr_key[i]][0] = 0

            count_match[0] += 1

        if reward == -1:
            if count_turn[0] < CHAIN_LENGTH:
                action = np.argmax(validActions * temp_bias[-1])
                action_history[count_turn[0]] = action
            else:
                key = encode(action_history)
                if key not in count_bias:
                    count_bias[key] = np.full(1, 0, dtype=np.int64)
                    per_bias[key] = np.full(env.getActionSize(), 0, dtype=np.int64) + 1
                    temp_bias[key] = np.arange(env.getActionSize(), dtype=np.int64) + 1
                    np.random.shuffle(temp_bias[key])

                action = np.argmax(validActions * temp_bias[key])
                action_history[0 : CHAIN_LENGTH - 1] = action_history[1:CHAIN_LENGTH]
                action_history[CHAIN_LENGTH - 1] = action
                chain_history[count_turn[0] - CHAIN_LENGTH] = key

            count_turn[0] += 1
        else:
            if reward == 1:
                per_bias[-1] += temp_bias[-1]
                for i in range(count_turn[0] - CHAIN_LENGTH):
                    per_bias[chain_history[i]] += temp_bias[chain_history[i]]
                    count_bias[chain_history[i]][0] += 1
            else:
                np.random.shuffle(temp_bias[-1])
                for i in range(count_turn[0] - CHAIN_LENGTH):
                    np.random.shuffle(temp_bias[chain_history[i]])
                    count_bias[chain_history[i]][0] += 1

            count_match[0] += 1
            count_turn[0] = 0
            action_history[:] = 0
            chain_history[:] = -1

            action = np.argmax(validActions * temp_bias[-1])

        return action, perData

    elif count_match[0] < 100000:
        if reward == -1:
            if count_turn[0] < CHAIN_LENGTH:
                action = np.argmax(validActions * temp_bias[-1])
                action_history[count_turn[0]] = action
            else:
                key = encode(action_history)
                if key not in count_bias:
                    action = np.argmax(validActions * temp_bias[-1])
                    check = False
                else:
                    action = np.argmax(validActions * temp_bias[key])
                    check = True

                action_history[0 : CHAIN_LENGTH - 1] = action_history[1:CHAIN_LENGTH]
                action_history[CHAIN_LENGTH - 1] = action
                if check:
                    chain_history[count_turn[0] - CHAIN_LENGTH] = key

            count_turn[0] += 1
        else:
            if reward == 1:
                per_bias[-1] += temp_bias[-1]
                for i in range(count_turn[0] - CHAIN_LENGTH):
                    if chain_history[i] == -1:
                        break

                    per_bias[chain_history[i]] += temp_bias[chain_history[i]]
                    count_bias[chain_history[i]][0] += 1
            else:
                np.random.shuffle(temp_bias[-1])
                for i in range(count_turn[0] - CHAIN_LENGTH):
                    if chain_history[i] == -1:
                        break

                    np.random.shuffle(temp_bias[chain_history[i]])
                    count_bias[chain_history[i]][0] += 1

            count_match[0] += 1
            count_turn[0] = 0
            action_history[:] = 0
            chain_history[:] = -1
            if count_match[0] == 100000:
                count_match[0] = -1

            action = np.argmax(validActions * temp_bias[-1])

        return action, perData


@njit()
def Test(state, perData):
    per_bias = perData[2]
    action_history = perData[0][1]
    count_turn = perData[0][2]

    reward = env.getReward(state)
    validActions = env.getValidActions(state)

    if reward == -1:
        if count_turn[0] < CHAIN_LENGTH:
            action = np.argmax(validActions * per_bias[-1] + validActions)
            action_history[count_turn[0]] = action
        else:
            key = encode(action_history)
            if key not in per_bias:
                action = np.argmax(validActions * per_bias[-1] + validActions)
            else:
                action = np.argmax(validActions * per_bias[key] + validActions)

            action_history[0 : CHAIN_LENGTH - 1] = action_history[1:CHAIN_LENGTH]
            action_history[CHAIN_LENGTH - 1] = action

        count_turn[0] += 1
    else:
        count_turn[0] = 0
        action_history[:] = 0

        action = np.argmax(validActions * per_bias[-1] + validActions)

    return action, perData


def convert_to_save(perData):
    if 1 not in perData.keys():
        raise Exception("Data này đã được convert rồi.")
    temp = dict()
    for key in perData.keys():
        temp[key] = dict()
        for key1 in perData[key].keys():
            temp[key][key1] = perData[key][key1]

    data = dict()
    data[0] = temp[0]
    data[0].pop(0)
    data[0].pop(3)
    for key in data[0].keys():
        data[0][key] = data[0][key].astype(np.int16)

    data[2] = temp[2]
    for key in data[2].keys():
        data[2][key] = np.argsort(np.argsort(data[2][key])).astype(np.int16)

    return data


def convert_to_test(perSave):
    perSave = perSave[()]
    temp = Dict()
    for key in perSave.keys():
        temp1 = Dict()
        for key1 in perSave[key].keys():
            temp1[key1] = perSave[key][key1]

        temp[key] = temp1

    return temp



import numpy as np
from numba import njit
from src.Utils import load_module_player


@njit()
def initEnv(n):
    """
    n : số lượng người chơi\n
    id| Name card    | Amount\n
    0 | Tempura      | 14\n
    1 | Sashimi      | 14\n
    2 | Dumpling     | 14\n
    3 | 1 Maki Roll  | 6\n
    4 | 2 Maki Roll  | 12\n
    5 | 3 Maki Roll  | 8\n
    6 | Salmon Nigiri| 10\n
    7 | Squid Nigiri | 5\n
    8 | Egg Nigiri   | 5\n
    9| Pudding      | 10\n
    10| Wasabi       | 6\n
    11| Chopsticks   | 4\n
    """
    amount_card = np.array([14, 14, 14, 6, 12, 8, 10, 5, 5, 10, 6, 4])
    id_card = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11])
    state_sys = np.array([1, 1, n])
    list_card = np.array([0])
    index = 0
    for amount in amount_card:
        list_c = np.array([id_card[index] for i in range(amount)])
        if len(list_card) == 1:
            list_card = list_c
        else:
            list_card = np.concatenate((list_card, list_c))
        index += 1
    np.random.shuffle(list_card)
    cards = list_card[: 3 * n * (12 - n)]
    state_sys = np.concatenate((state_sys, cards))
    for i in range(n):
        state_sys = np.concatenate(
            (state_sys, np.array([0, 0]), np.array([-1 for i in range(12 - n)]))
        )
    return state_sys


@njit()
def getAgentState(state_sys, index_player):
    amount_player = 5
    round = state_sys[0] - 1
    turn = state_sys[1] + 4
    index_start_card_board = (
        (((turn - index_player) % amount_player)) * 7 + round * 7 * amount_player + 3
    )
    index_end_card_board = index_start_card_board + 7
    state_player = np.zeros(86)
    state_player[0] = state_sys[0]
    state_player[1] = state_sys[1]
    state_player[84:86] = np.array([0, 1])
    step = 2
    for index_player_relative in range(index_player, index_player + amount_player):
        index_start_player_s = (
            (index_player_relative % amount_player) * 7
            + 3 * amount_player * 7
            + (index_player_relative % amount_player + 1) * 2
            + 1
        )
        index_end_player_s = index_start_player_s + 9
        if index_player == index_player_relative:
            for i in state_sys[index_start_card_board:index_end_card_board]:
                if i != -1:
                    state_player[step + i] += 1
            #  print("cover:",index_player,state_sys,state_sys[index_start_card_board:index_end_card_board],state_player[step:step+12])
            step += 12
            state_player[step : step + 2] = state_sys[
                index_start_player_s : index_start_player_s + 2
            ]
            step += 2
            for i in state_sys[index_start_player_s + 2 : index_end_player_s]:
                if i != -1:
                    state_player[step + i] += 1
            step += 12
        else:
            state_player[step : step + 2] = state_sys[
                index_start_player_s : index_start_player_s + 2
            ]
            step += 2
            for i in state_sys[index_start_player_s + 2 : index_end_player_s]:
                if i != -1:
                    state_player[step + i] += 1
            step += 12
    return state_player.astype(np.float64)


@njit()
def calculator_for_one(card):
    score_dumpling = [0, 1, 3, 6, 10, 15]
    score = card[0]
    card = card[2:]
    #  print("card",card,end = " ")
    tempura = np.count_nonzero(card == 0)

    score += tempura // 2 * 5
    #  print("tempura: ",score,end = " ")

    sashimi = np.count_nonzero(card == 1)
    score += sashimi // 3 * 10
    #  print("sashimi: ",score,end = " ")

    dumpling = np.count_nonzero(card == 2)
    if dumpling > 5:
        score += score_dumpling[dumpling - 5]
        dumpling = 5
    score += score_dumpling[dumpling]

    #  print("dumpling: ",score,end = " ")

    salmon_nigiri = np.count_nonzero(card == 6)
    squid_nigiri = np.count_nonzero(card == 7)
    egg_nigiri = np.count_nonzero(card == 8)
    wasabi = np.count_nonzero(card == 10)
    list_ = [squid_nigiri, salmon_nigiri, egg_nigiri]
    for i in range(len(list_)):
        for j in range(list_[i]):
            if wasabi != 0:
                score += (3 - i) * 3
                wasabi -= 1
            else:
                score += 3 - i

    return score, np.count_nonzero(card == 9)


@njit()
def count_maki(card):
    maki_1 = np.count_nonzero(card == 3)
    maki_2 = np.count_nonzero(card == 4)
    maki_3 = np.count_nonzero(card == 5)
    return maki_1 + maki_2 * 2 + maki_3 * 3


@njit()
def get_index(arr, first, second):
    return np.where(arr == first)[0], np.where(arr == second)[0]


@njit()
def calculator_pudding(state_sys, amount_player):
    amount_player = state_sys[2]
    arr_pudding = np.array([0 for i in range(amount_player)])
    for index_player_relative in range(0, amount_player):
        index_start_player_s = (
            (index_player_relative) * 7
            + 3 * amount_player * 7
            + (index_player_relative + 1) * 2
            + 1
        )
        pudding = state_sys[index_start_player_s + 1]
        arr_pudding[index_player_relative] = pudding

    max_p, min_p = max(arr_pudding), min(arr_pudding)
    list_ = get_index(arr_pudding, max_p, min_p)
    #  print(list_)
    for top in range(len(list_)):
        for index_player_relative in list_[top]:
            score = (6 - top * 12) // len(list_[top])
            index_start_player_s = (
                (index_player_relative) * 7
                + 3 * amount_player * 7
                + (index_player_relative + 1) * 2
                + 1
            )
            state_sys[index_start_player_s] += score
    return state_sys


@njit()
def calculator_score(state_sys, amount_player):
    amount_player = state_sys[2]
    round = state_sys[0] - 1
    arr_maki = np.array([0 for i in range(amount_player)])
    first, second = -1, -1
    if round != 4:
        for index_player_relative in range(0, amount_player):
            index_start_player_s = (
                (index_player_relative) * 7
                + 3 * amount_player * 7
                + (index_player_relative + 1) * 2
                + 1
            )
            index_end_player_s = (
                (index_player_relative + 1) * 7
                + 3 * amount_player * 7
                + (index_player_relative + 2) * 2
                + 1
            )
            card = state_sys[index_start_player_s:index_end_player_s]
            state_sys[index_start_player_s], puding = calculator_for_one(card)
            state_sys[index_start_player_s + 1] += puding
            c_maki = count_maki(card)
            arr_maki[index_player_relative] = c_maki
            if c_maki > first:
                second = first
                first = c_maki
            elif c_maki > second:
                second = c_maki

        list_ = get_index(arr_maki, first, second)
        #  print(list_)
        for top in range(len(list_)):
            for index_player_relative in list_[top]:
                score = (6 - top * 3) // len(list_[top])
                index_start_player_s = (
                    (index_player_relative) * 7
                    + 3 * amount_player * 7
                    + (index_player_relative + 1) * 2
                    + 1
                )
                state_sys[index_start_player_s] = (
                    state_sys[index_start_player_s] + score
                )
                #  print(score)
            if len(list_[top]) > 1:
                break
    return state_sys


@njit()
def getReward(state_player):
    if state_player[1] <= 21:
        return -1
    list_score = state_player[14::14]
    Max_Score = max(list_score)
    list_winner = np.where(list_score == Max_Score)[0]
    check_winer_self = np.where(list_winner == 0)[0]
    if len(check_winer_self) == 1:
        if len(list_winner) == 1:
            return 1
        else:
            list_pudding = state_player[15::14]
            pudding_victoryer = list_pudding[list_winner]
            max_puding = max(pudding_victoryer)
            if list_pudding[0] == max_puding:
                return 1
            else:
                return 0
    else:
        return 0


@njit()
def winner_victory(state_sys):
    amount_player = int(state_sys[2])
    list_score = state_sys[3 + 3 * amount_player * 7 :: 14 - amount_player]
    max_score = max(list_score)
    winner = np.where(list_score == max_score)[0]
    #  print("Diem:",list_score)
    if len(winner) == 1:
        return winner
    else:
        list_pudding = state_sys[3 + 3 * amount_player * 7 + 1 :: 14 - amount_player]
        pudding_victoryer = list_pudding[winner]
        max_puding = max(pudding_victoryer)
        winner_puding = np.where(pudding_victoryer == max_puding)[0]
        return winner[winner_puding]


@njit()
def stepEnv(state_sys, list_action, amount_player, turn, round):
    player = 0
    turn += 4
    for a in range(turn + amount_player, turn, -1):
        index_board_s = (
            (a % amount_player) * (12 - amount_player)
            + round * (12 - amount_player) * amount_player
            + 3
        )
        index_board_e = index_board_s + (12 - amount_player)
        index_player_s = (
            (player % amount_player) * (12 - amount_player)
            + 3 * amount_player * (12 - amount_player)
            + (player % amount_player + 1) * 2
            + 3
        )
        index_player_e = index_player_s + (12 - amount_player)
        l_a = list_action[player]
        l_a = l_a[np.where(l_a >= 0)[0]]
        for i in l_a:
            if i == 13:
                break
            #  print("----------------------",i,state_sys[index_player_s:index_player_e],state_sys[index_board_s:index_board_e],state_sys.shape)
            if i == 12:
                state_sys = move_card_step(
                    state_sys,
                    11,
                    index_player_s,
                    index_player_e,
                    index_board_s,
                    index_board_e,
                )
                continue
            state_sys = move_card_step(
                state_sys,
                i,
                index_board_s,
                index_board_e,
                index_player_s,
                index_player_e,
            )
        player += 1
    return state_sys


@njit()
def reset_card_player(state_sys):
    amount_player = state_sys[2]
    for player in range(amount_player):
        index_player_s = (
            (player % amount_player) * (12 - amount_player)
            + 3 * amount_player * 7
            + (player % amount_player + 1) * 2
            + 3
        )
        index_player_e = index_player_s + (12 - amount_player)
        for i in range(index_player_s, index_player_e):
            state_sys[i] = -1
    return state_sys


@njit()
def test_action(player_state, action):
    index_between = 14
    if action == 12:
        player_state = move_card(player_state, 11, index_between + 2, 2)
        player_state[-2] = 1
        return player_state
    #  player_state[-2] = 0
    if player_state[-1] > 0:
        player_state = move_card(player_state, action, 2, index_between + 2)
        player_state[-1] -= 1
    else:
        player_state = move_card(player_state, action, 2, index_between + 2)
        player_state[-2] -= 1
    return player_state


@njit()
def move_card(state, card, start_1=0, start_2=0):
    state[start_1 + card] -= 1
    state[start_2 + card] += 1
    return state


@njit()
def move_card_step(state, card, start_1=0, end_1=0, start_2=0, end_2=0):
    index_relative_from = np.where(state[start_1:end_1] == card)[0]
    index_relative_to = np.where(state[start_2:end_2] == -1)[0]
    index_relative_from = index_relative_from[0] + start_1
    index_relative_to = index_relative_to[0] + start_2
    temp = state[index_relative_from]
    state[index_relative_from] = state[index_relative_to]
    state[index_relative_to] = temp
    return state


@njit()
def getActionSize():
    return 14


@njit()
def getAgentSize():
    return 5


@njit()
def getStateSize():
    return 86


@njit()
def getValidActions(player_state_origin: np.int64):
    list_action_return = np.zeros(14)
    player_state = player_state_origin.copy()
    player_state = player_state.astype(np.int64)
    amount = 5
    card = player_state[2:14]
    #  print("card:", card)
    if player_state[-2] == 1:
        card[11] -= 1
    list_action = np.where(card > 0)[0]
    if (12 - amount) * 3 < player_state[1]:
        list_action = np.array([13])

    of_card = player_state[16:28]
    if of_card[11] != 0 and player_state[-2] != 1 and sum(card) > 1:
        list_action = np.append(list_action, np.array([12]))

    list_action_return[np.unique(list_action)] = 1

    return list_action_return.astype(np.float64)


@njit()
def one_game_numba(p0, list_other, per_player, per1, per2, per3, per4, p1, p2, p3, p4):
    amount_player = 5
    state_sys = initEnv(amount_player)
    amount_player = state_sys[2]
    turn = state_sys[1]

    while turn < 7 * 3:
        round = state_sys[0] - 1
        turn = state_sys[1]
        list_action = np.full((amount_player, 3), 13)
        for idx in range(amount_player):
            player_state = getAgentState(state_sys, idx)

            count = 0
            while player_state[-1] + player_state[-2] > 0:
                if list_other[idx] == -1:
                    action, per_player = p0(player_state, per_player)
                    lst_action = getValidActions(player_state)
                    if lst_action[action] != 1:
                        raise Exception("Action không hợp lệ")
                elif list_other[idx] == 1:
                    action, per1 = p1(player_state, per1)
                elif list_other[idx] == 2:
                    action, per2 = p2(player_state, per2)
                elif list_other[idx] == 3:
                    action, per3 = p3(player_state, per3)
                elif list_other[idx] == 4:
                    action, per4 = p4(player_state, per4)
                list_action[idx][count] = action
                count += 1
                player_state = test_action(player_state, action)
            #  player_state = getAgentState(state_sys,idx)
        state_sys = stepEnv(state_sys, list_action, amount_player, turn, round)
        if turn % 7 == 0:
            state_sys = calculator_score(state_sys, amount_player)
            if state_sys[0] < 3:
                state_sys[0] += 1
                state_sys = reset_card_player(state_sys)
        if turn == 7 * 3:
            state_sys = calculator_pudding(state_sys, amount_player)
        if turn <= 7 * 3:
            state_sys[1] += 1

    for idx in range(amount_player):
        player_state = getAgentState(state_sys, idx)
        if list_other[idx] == -1:
            action, per_player = p0(player_state, per_player)
        elif list_other[idx] == 1:
            action, per1 = p1(player_state, per1)
        elif list_other[idx] == 2:
            action, per2 = p2(player_state, per2)
        elif list_other[idx] == 3:
            action, per3 = p3(player_state, per3)
        elif list_other[idx] == 4:
            action, per4 = p4(player_state, per4)
    winner = winner_victory(state_sys)
    p0_idx = np.where(list_other == -1)[0][0]
    if p0_idx in winner:
        result = 1
    else:
        result = 0
    return result, per_player


@njit()
def n_games_numba(
    p0, num_game, per_player, list_other, per1, per2, per3, per4, p1, p2, p3, p4
):
    win = 0
    for _ in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_numba(
            p0, list_other, per_player, per1, per2, per3, per4, p1, p2, p3, p4
        )
        win += winner

    return win, per_player


def one_game_normal(p0, list_other, per_player, per1, per2, per3, per4, p1, p2, p3, p4):
    amount_player = 5
    state_sys = initEnv(amount_player)
    amount_player = state_sys[2]
    turn = state_sys[1]

    while turn < 7 * 3:
        round = state_sys[0] - 1
        turn = state_sys[1]
        list_action = np.full((amount_player, 3), 13)
        for idx in range(amount_player):
            player_state = getAgentState(state_sys, idx)

            count = 0
            while player_state[-1] + player_state[-2] > 0:
                if list_other[idx] == -1:
                    action, per_player = p0(player_state, per_player)
                    lst_action = getValidActions(player_state)
                    if lst_action[action] != 1:
                        raise Exception("Action không hợp lệ")
                elif list_other[idx] == 1:
                    action, per1 = p1(player_state, per1)
                elif list_other[idx] == 2:
                    action, per2 = p2(player_state, per2)
                elif list_other[idx] == 3:
                    action, per3 = p3(player_state, per3)
                elif list_other[idx] == 4:
                    action, per4 = p4(player_state, per4)
                list_action[idx][count] = action
                count += 1
                player_state = test_action(player_state, action)
            player_state = getAgentState(state_sys, idx)
        state_sys = stepEnv(state_sys, list_action, amount_player, turn, round)
        if turn % 7 == 0:
            state_sys = calculator_score(state_sys, amount_player)
            if state_sys[0] < 3:
                state_sys[0] += 1
                state_sys = reset_card_player(state_sys)
        if turn == 7 * 3:
            state_sys = calculator_pudding(state_sys, amount_player)
        if turn <= 7 * 3:
            state_sys[1] += 1

    for idx in range(amount_player):
        player_state = getAgentState(state_sys, idx)
        if list_other[idx] == -1:
            action, per_player = p0(player_state, per_player)
        elif list_other[idx] == 1:
            action, per1 = p1(player_state, per1)
        elif list_other[idx] == 2:
            action, per2 = p2(player_state, per2)
        elif list_other[idx] == 3:
            action, per3 = p3(player_state, per3)
        elif list_other[idx] == 4:
            action, per4 = p4(player_state, per4)
    winner = winner_victory(state_sys)
    p0_idx = np.where(list_other == -1)[0][0]
    if p0_idx in winner:
        result = 1
    else:
        result = 0
    return result, per_player


def n_games_normal(
    p0, num_game, per_player, list_other, per1, per2, per3, per4, p1, p2, p3, p4
):
    win = 0
    for _ in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_normal(
            p0, list_other, per_player, per1, per2, per3, per4, p1, p2, p3, p4
        )
        win += winner

    return win, per_player


import json, sys

try:
    from env import SHORT_PATH
except:
    pass


@njit()
def bot_lv0(state, perData):
    validActions = getValidActions(state)
    arr_action = np.where(validActions == 1)[0]
    idx = np.random.randint(0, arr_action.shape[0])
    return arr_action[idx], perData


@njit()
def check_run_under_njit(agent, perData):
    return True


def load_agent(level, *args):
    num_bot = getAgentSize() - 1

    if "_level_" not in globals():
        global _level_
        _level_ = level
        init = True
    else:
        if _level_ != level:
            _level_ = level
            init = True
        else:
            init = False

    if init:
        global _list_per_level_
        global _list_bot_level_
        _list_per_level_ = []
        _list_bot_level_ = []

        if _level_ == 0:
            _list_per_level_ = [
                np.array([[0.0]], dtype=np.float64) for _ in range(num_bot)
            ]
            _list_bot_level_ = [bot_lv0 for _ in range(num_bot)]
        else:
            env_name = sys.argv[1]
            if len(args) > 0:
                dict_level = json.load(
                    open(f"{SHORT_PATH}src/Log/check_system_about_level.json")
                )
            else:
                dict_level = json.load(open(f"{SHORT_PATH}src/Log/level_game.json"))

            if str(_level_) not in dict_level[env_name]:
                raise Exception("Hiện tại không có level này")

            lst_agent_level = dict_level[env_name][str(level)][2]

            for i in range(num_bot):
                if level == -1:
                    module_agent = load_module_player(
                        lst_agent_level[i], game_name=env_name
                    )
                    _list_per_level_.append(module_agent.DataAgent())
                else:
                    data_agent_level = np.load(
                        f"{SHORT_PATH}src/Agent/{lst_agent_level[i]}/Data/{env_name}_{level}/Train.npy",
                        allow_pickle=True,
                    )
                    module_agent = load_module_player(lst_agent_level[i])
                    _list_per_level_.append(
                        module_agent.convert_to_test(data_agent_level)
                    )
                _list_bot_level_.append(module_agent.Test)

    return _list_bot_level_, _list_per_level_


def run(
    p0: any = bot_lv0,
    num_game: int = 100,
    per_player: np.ndarray = np.array([[0.0]]),
    level: int = 0,
    *args,
):
    num_bot = getAgentSize() - 1
    list_other = np.array([-1] + [i + 1 for i in range(num_bot)])
    try:
        check_njit = check_run_under_njit(p0, per_player)
    except:
        check_njit = False

    load_agent(level, *args)

    if check_njit:
        return n_games_numba(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_per_level_[3],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
            _list_bot_level_[3],
        )
    else:
        return n_games_normal(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_per_level_[3],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
            _list_bot_level_[3],
        )



import warnings

import numpy as np
from numba import njit
from numba.core.errors import (
    NumbaDeprecationWarning,
    NumbaExperimentalFeatureWarning,
    NumbaPendingDeprecationWarning,
    NumbaWarning,
)

import env

warnings.simplefilter("ignore", category=NumbaDeprecationWarning)
warnings.simplefilter("ignore", category=NumbaPendingDeprecationWarning)
warnings.simplefilter("ignore", category=NumbaExperimentalFeatureWarning)
warnings.simplefilter("ignore", category=NumbaWarning)

COUNT_TEST = 10000


def CheckAllFunc(BOOL_CHECK_ENV, msg):
    for func in [
        "getActionSize",
        "getStateSize",
        "getAgentSize",
        "getReward",
        "getValidActions",
        "run",
    ]:
        try:
            getattr(func)
        except:
            msg.append(f"Không có hàm: {func}")
            BOOL_CHECK_ENV = False
    return BOOL_CHECK_ENV


def CheckReturn(BOOL_CHECK_ENV, msg):
    for func in ["getActionSize", "getStateSize", "getAgentSize"]:
        try:
            func_ = getattr(func)
            out = func_()
            if type(out) != int and type(out) != np.int64:
                msg.append(
                    f"ham {func} tra sai dau ra: dau ra yeu cau int, dau ra hien tai: {type(out)}"
                )
                BOOL_CHECK_ENV = False
        except:
            pass
    return BOOL_CHECK_ENV


def RunGame(BOOL_CHECK_ENV, msg):
    @njit()
    def test_numba(p_state, per_file):
        arr_action = env.getValidActions(p_state)

        if env.getReward(p_state) != -1:
            per_file[0] += 1
        if env.getReward(p_state) == 1:
            per_file[1] += 1
        if np.min(p_state) < 0:
            per_file[2] = 1
        if len(p_state) != env.getStateSize():
            per_file[3] = 1
        if len(arr_action) != env.getActionSize():
            per_file[4] = 1

        arr_action = np.where(arr_action == 1)[0]
        act_idx = np.random.randint(0, len(arr_action))
        return arr_action[act_idx], per_file

    def test_no_numba(p_state, per_file):
        arr_action = env.getValidActions(p_state)
        if p_state.dtype != np.float64:
            per_file[5] = 1
        if arr_action.dtype != np.float64:
            per_file[6] = 1
        if np.where((arr_action != 1) & (arr_action != 0))[0].shape[0] > 0:
            per_file[7] = 1
        arr_action = np.where(arr_action == 1)[0]
        act_idx = np.random.randint(0, len(arr_action))
        return arr_action[act_idx], per_file

    try:
        per = np.array(
            [0, 0, 0, 0, 0, 0, 0, 0]
        )  # end, win end, state âm, state thay đổi, actions thay đổi, type state, type action
        win, per = env.run(test_numba, COUNT_TEST, per, 0)

        if type(win) != int and type(win) != np.int64:
            msg.append("hàm run trả ra sai đầu ra, yêu cầu int")
            BOOL_CHECK_ENV = False
        if per[0] != COUNT_TEST:
            msg.append(f"run, Số trận kết thúc khác với số trận test, {per[0]}")
            BOOL_CHECK_ENV = False
        if per[1] != win:
            msg.append(
                f"run, Số trận thắng khi kết thúc khác với sô trận check. Thắng khi dùng getReward: {per[1]}, win: {win}"
            )
            BOOL_CHECK_ENV = False
        if per[2] == 1:
            msg.append(f"State đang bị âm")
            BOOL_CHECK_ENV = False
        if per[3] == 1:
            msg.append(f"len STATE đang thay đổi trong quá trình chạy game")
            BOOL_CHECK_ENV = False
        if per[4] == 1:
            msg.append(f"len ACTION đang thay đổi trong quá trình chạy game")
            BOOL_CHECK_ENV = False

        try:
            per = np.array([0, 0, 0, 0, 0, 0, 0, 0])
            win, per = env.run(test_no_numba, COUNT_TEST, per, 0)
            if per[5] == 1:
                msg.append(f"STATE đang trả ra sai output")
                BOOL_CHECK_ENV = False
            if per[6] == 1:
                msg.append(f"array ACTION đang trả ra sai định dạng output")
                BOOL_CHECK_ENV = False
            if per[7] == 1:
                msg.append(f"array ACTION đang trả ra sai output, != 0, 1")
                BOOL_CHECK_ENV = False
        except:
            msg.append("hàm run không train được với agent không numba")
            BOOL_CHECK_ENV = False

    except:
        msg.append(f"hàm run đang bị lỗi")
        BOOL_CHECK_ENV = False

    return BOOL_CHECK_ENV


def CheckRunGame(BOOL_CHECK_ENV, msg):
    BOOL_CHECK_ENV = RunGame(BOOL_CHECK_ENV, msg)
    return BOOL_CHECK_ENV


def CheckRandomState(BOOL_CHECK_ENV, msg):
    for i in range(10000):
        p_state1 = np.random.randint(-100, 100, env.getStateSize())
        p_state2 = np.random.randint(0, 1000, env.getStateSize())
        p_state3 = np.random.randn(env.getStateSize()) * 100
        try:
            env.getValidActions(p_state1)
            env.getValidActions(p_state2)
            env.getValidActions(p_state3)
        except:
            msg.append(f"hàm getValidActions lỗi khi nhận vào random state")
            BOOL_CHECK_ENV = False
            return BOOL_CHECK_ENV
        try:
            env.getReward(p_state1)
            env.getReward(p_state2)
            env.getReward(p_state3)
        except:
            msg.append(f"hàm getReward lỗi khi nhận vào random state")
            BOOL_CHECK_ENV = False
            return BOOL_CHECK_ENV
    try:
        env.getReward(np.full(env.getStateSize(), 0))
        env.getReward(np.full(env.getStateSize(), 0.0))
        env.getReward(np.full(env.getStateSize(), 1))
    except:
        msg.append(f"hàm getReward lỗi khi nhận vào random state")
        BOOL_CHECK_ENV = False
        return BOOL_CHECK_ENV
    try:
        env.getValidActions(np.full(env.getStateSize(), 0))
        env.getValidActions(np.full(env.getStateSize(), 0.0))
        env.getValidActions(np.full(env.getStateSize(), 1))
    except:
        msg.append(f"hàm getValidActions lỗi khi nhận vào random state")
        BOOL_CHECK_ENV = False
        return BOOL_CHECK_ENV
    return BOOL_CHECK_ENV


def check_env():
    BOOL_CHECK_ENV = True
    msg = []
    BOOL_CHECK_ENV = CheckAllFunc(BOOL_CHECK_ENV, msg)
    BOOL_CHECK_ENV = CheckReturn(BOOL_CHECK_ENV, msg)
    BOOL_CHECK_ENV = CheckRunGame(BOOL_CHECK_ENV, msg)
    BOOL_CHECK_ENV = CheckRandomState(BOOL_CHECK_ENV, msg)
    return BOOL_CHECK_ENV, msg


def check_pytest(env_str):
    env.make(env_str)
    BOOL_CHECK_ENV, msg = check_env()
    return BOOL_CHECK_ENV, msg



import sys

import numba
import numpy as np
from numba import njit
from numba.typed import List


###########################
@njit
def getActionSize():
    return 62


@njit
def getStateSize():
    return 103


@njit
def initEnv():
    """init env"""
    env = np.zeros(87)
    card = np.arange(46)  # card except Defuse and Explo kitten
    np.random.shuffle(card)
    env[:56] += 5  #  5 is card on draw pile
    for i in range(5):  #  draw 4 card for player: id from 0 to 4
        env[card[i * 4 : (i + 1) * 4]] = i
        env[46 + i] = i

    draw_pile = np.where(env == 5.0)[0].astype(np.float64)
    np.random.shuffle(draw_pile)

    discard_pile = np.zeros(13)  # card on discard pile will have id 6

    env[56] = 0  #  nope count
    env[57] = 0  #  track player id main turn
    env[58:62] = [2, 3, 4, 0]  #  track player id Nope turn
    env[62:67] = 1  #  0 if lose else 1
    env[
        67
    ] = 0  # phase [0:main turn, 1:nope turn,2:steal card turn,3:choose/take card turn, 4: discard turn]
    env[68] = 1  #  number of card player env[57] have to draw
    env[69:72] = [-1, -1, -1]  # three card in see the future
    env[72] = -1  #  player env[57] last action
    env[73] = env[57] + 1  # player id in nope turn
    env[74] = -1  # player chosen in phase 2

    env[75] = 1  # num card main player have to discard
    env[76:87] = 0
    return env, draw_pile, discard_pile


@njit
def getNumCard(env, idx):
    """Get the number of card with given type"""
    return np.where(env == idx)[0].shape[0]


@njit
def getAllNumCard(env, idx):
    """Get all the number of card"""
    state = np.zeros(12)
    state[0] = getNumCard(env[0:5], idx)
    for i in range(4):
        state[1 + i] = getNumCard(env[5 + i * 4 : 9 + i * 4], idx)
    state[5] = getNumCard(env[21:26], idx)
    for i in range(5):
        state[6 + i] = getNumCard(env[26 + i * 4 : 30 + i * 4], idx)
    state[11] = getNumCard(env[46:52], idx)
    return state


@njit
def getCardType(idx):
    """Get the type of the card"""
    cards = List(
        [
            np.arange(0.0, 5.0),
            np.arange(5.0, 9.0),
            np.arange(9.0, 13.0),
            np.arange(13.0, 17.0),
            np.arange(17.0, 21.0),
            np.arange(21.0, 26.0),
            np.arange(26.0, 30.0),
            np.arange(30.0, 34.0),
            np.arange(34.0, 38.0),
            np.arange(38.0, 42.0),
            np.arange(42.0, 46.0),
            np.arange(46.0, 52.0),
            np.arange(52.0, 56.0),
        ]
    )
    i = 0
    for c in cards:
        if idx in c:
            return i
        else:
            i += 1


@njit
def getCardRange(type_card):
    """Get the range of the card given type"""
    cards = List(
        [
            np.arange(0.0, 5.0),
            np.arange(5.0, 9.0),
            np.arange(9.0, 13.0),
            np.arange(13.0, 17.0),
            np.arange(17.0, 21.0),
            np.arange(21.0, 26.0),
            np.arange(26.0, 30.0),
            np.arange(30.0, 34.0),
            np.arange(34.0, 38.0),
            np.arange(38.0, 42.0),
            np.arange(42.0, 46.0),
            np.arange(46.0, 52.0),
            np.arange(52.0, 56.0),
        ]
    )

    return (
        cards[type_card].astype(np.int64)[0],
        cards[type_card].astype(np.int64)[-1] + 1,
    )


@njit
def getAgentState(env, draw_pile, discard_pile):
    state = np.zeros(getStateSize())
    # get card
    phase = env[67]
    main_id = env[57]
    nope_id = env[73]
    last_action = env[72]
    if phase == 0:
        pIdx = int(main_id)
    elif phase == 1:
        pIdx = int(nope_id)
    elif phase == 2:
        pIdx = int(main_id)
    elif phase == 3:
        if last_action == 3:
            pIdx = int(env[74])
        else:
            pIdx = int(main_id)
    elif phase == 4:
        pIdx = int(main_id)
    state[0:12] = getAllNumCard(env, pIdx)
    state[12:25] = discard_pile  # discard pile
    state[25] = np.where(draw_pile != -1)[0].shape[0]  # number of card in draw pile
    state[26] = np.where(env[62:67] == 1)[0].shape[0]
    state[27] = env[56] % 2  # 1 if action been Nope else 0

    if phase != 4:  # not discard turn
        state[67:71][int(phase)] = 1  # phase

    if last_action >= 0:
        state[72:82][int(last_action)] = 1  #  player main turn last action

    for i in range(4):  # number of card other people have
        state[87 + i] = np.where(env[0:56] == env[58 + i])[0].shape[0]

    if phase == 1:  # nope turn
        nope_turn = nopeTurn(nope_id)
        for i in range(4):
            state[82 + i] = env[62:67][int(nope_turn[i])]
        state[86] = env[62:67][int(nope_id)]  # lose or not

    elif phase == 3 and last_action == 3:  # choose / take card turn, action favor
        nope_turn = nopeTurn(env[74])
        for i in range(4):
            state[82 + i] = env[62:67][int(nope_turn[i])]
        state[86] = env[62:67][int(env[74])]  # lose or not

    else:
        if phase == 4:  # discard phase
            state[91:102] = env[76:87]  # card have been discard (5 combo)
            state[102] = env[75]  #  num card left have to discard

        if pIdx == int(main_id):
            for i in range(3):
                if env[69 + i] != -1:
                    card = np.zeros(13)
                    card[int(getCardType(env[69 + i]))] = 1
                    state[
                        28 + 13 * i : 41 + 13 * i
                    ] = card  #  three card if use see the future
        state[71] = env[68]  #  number of card player have to draw

        for i in range(4):
            state[82 + i] = env[62:67][int(env[58 + i])]
        state[86] = env[62:67][int(env[57])]  # lose or not

    return state


@njit
def getValidActions(state):
    list_action = np.zeros(getActionSize())
    available_card = (state[0:11] > 0).astype(np.float64)
    if state[67] == 1:  # main turn
        list_action[1:6] = (state[1:6] > 0).astype(np.float64)
        if np.sum(state[87:91]) == 0:
            list_action[3] = 0
        list_action[6] = 1
        if np.max(state[0:11]) >= 2 and np.sum(state[87:91]) > 0:  # two of a kind
            list_action[7] = True
        if np.max(state[0:11]) >= 3 and np.sum(state[87:91]) > 0:  # three of a kind
            list_action[8] = True

        if np.sum(available_card) >= 5 and np.sum(state[12:24]) > 0:  # five of a kind
            list_action[9] = True
        if np.sum(list_action) == 0:
            list_action[10] = 1
    elif state[68] == 1:  # Nope turn
        if state[0] > 0:
            list_action[0] = 1  # Nope
        list_action[10] = 1  # skip Nope

    elif state[69] == 1:  # steal turn
        for i in range(4):
            if state[82 + i] == 1 and state[87 + i] > 0:
                list_action[11 + i] = 1
        if np.sum(list_action[11:15]) == 0:
            list_action[6] = 1

    elif state[70] == 1:  # choose/take card turn
        main_action = np.where(state[72:82] == 1)[0][0]
        if main_action == 3:
            list_action[15:27][np.where(state[0:12] > 0)] = 1
        elif main_action == 8:
            list_action[27:39] = 1
        elif main_action == 9:
            list_action[39:50] = ((state[12:23] - state[91:102]) > 0).astype(
                np.float64
            )  # other
            list_action[50] = (state[23] > 0) * 1.0  # defuse

    elif sum(state[67:71]) == 0:  # discard turn
        last_action = np.argmax(state[72:82])
        if last_action == 7:
            available_card = (state[0:11] >= 2).astype(np.float64)
            list_action[51:62] = available_card
        elif last_action == 8:
            available_card = (state[0:11] >= 3).astype(np.float64)
            list_action[51:62] = available_card
        elif last_action == 9:
            list_action[51:62] = ((available_card - state[91:102]) > 0).astype(
                np.float64
            )

    #  if np.sum(list_action)==0:
    #      list_action[10] = 1
    return list_action


@njit
def checkDefuse(env, discard_pile):  #  get the Defuse (if player have else -1)
    card = np.where(env[46:52] == env[57])[0].astype(np.int64)
    if card.shape[0] > 0:
        card_id = card[0]
        env[46:52][card_id] = 6
        discard_pile[11] += 1
        #  print('Player ',env[57],' have Defuse!')
        return True
    return False


@njit
def checkExploding(card):  #  check if that card is expode or not
    explode = np.array([52, 53, 54, 55], dtype=np.float64)
    if card in explode:
        #  print('PLayer draw an Exploding kitten!')
        return True
    return False


@njit
def changeTurn(env, num_card_draw=1):
    """Change the main turn"""
    env[57] = int(env[57] + 1) % 5
    while env[62:67][int(env[57])] == 0:  # if player id is already lost.
        env[57] = int(env[57] + 1) % 5
    env[58:62] = nopeTurn(env[57])
    env[56] = 0  # reset nope count
    for i in range(4):
        if env[62:67][int(env[58:62][i])] == 1:
            env[73] = env[58:62][i]  # reset nope player id
    #  if env[68]:
    if env[68] >= 2:
        env[68] += num_card_draw  # card next player draw
    else:
        env[68] = num_card_draw
    env[67] = 0  #  change phase to 0
    env[69:72] = -1
    env[72] = -1  # reset last action
    return env


@njit
def drawCard(env, draw_pile, discard_pile):
    """Draw card"""
    #  print('Player ',env[57],' draw ', env[68],'card(s)')
    for i in range(int(env[68])):
        index_draw = np.where(draw_pile != -1)[0][0]
        #  print(f'Phase: {env[67]} Player {env[57]} Draw : {visualCard(np.array([draw_pile[index_draw]]))}')
        if checkExploding(draw_pile[index_draw]):  # draw an exploding kitten
            if checkDefuse(env, discard_pile):  # player have defuse
                idx = np.random.randint(index_draw, draw_pile.shape[0])
                draw_pile_2 = np.zeros_like(draw_pile)
                draw_pile_2[0:index_draw] = draw_pile[0:index_draw]
                draw_pile_2[index_draw:idx] = draw_pile[index_draw + 1 : idx + 1]
                draw_pile_2[idx] = draw_pile[index_draw]
                draw_pile_2[idx + 1 :] = draw_pile[idx + 1 :]
                draw_pile = draw_pile_2
                # insert explode card back to the Draw Pile
            else:  # player lost
                #  print('Player ',env[57],' loss!')
                env[62:67][int(env[57])] = 0
                env[0:56][np.where(env[0:56] == env[57])] = 6
                env[52:56][np.where(env[52:56] != 6)] = 6
                discard_pile[12] += 1
                draw_pile[index_draw] = -1
                break
        else:  # draw other card
            env[0:56][int(draw_pile[index_draw])] = env[57]  # draw
            draw_pile[index_draw] = -1
    env[68] = 0
    env = changeTurn(env, 1)
    return env, draw_pile, discard_pile


@njit
def nopeTurn(id):
    """Nope turn given the main player id"""
    return np.arange(id + 1.0, id + 5.0) % 5


@njit
def checkIfNope(env):
    """Return True if the main player's card has been Nope"""
    return env[56] % 2 == 1


@njit
def executeMainAction(env, draw_pile, discard_pile, action):
    """Execute main action if it has not been Nope"""
    #  print('Execute main Action!')
    env[56] = 0
    if action == 1:  # Attack
        #  print(f'Phase: {env[67]} Player {env[57]} attack!')
        env[67] = 0
        env = changeTurn(
            env, num_card_draw=2
        )  # change main turn, next player draw 2 card
    elif action == 2:  # Skip
        #  print(f'Phase: {env[67]} Player {env[57]} skip!')
        env[68] -= 1
        env[67] = 0
        if env[68] == 0:
            env = changeTurn(env, num_card_draw=1)
    elif action == 3:
        #  print(f'Phase: {env[67]} Player {env[57]} use favor!')
        env[67] = 2
    elif action == 4:  # Shuffle
        #  print(f'Phase: {env[67]} Player {env[57]} shuffle!')
        np.random.shuffle(draw_pile)
        env[67] = 0
        env[69:72] = -1  # reset future card
    elif action == 5:  # See the future
        #  print(f'Phase: {env[67]} Player {env[57]} see the future!')
        if np.where(draw_pile != -1)[0].shape[0] >= 3:
            env[69:72] = draw_pile[np.where(draw_pile != -1)[0][0:3]]
        else:
            env[69:72] = np.concatenate(
                (
                    draw_pile[np.where(draw_pile != -1)[0][0:3]],
                    np.zeros(3 - np.where(draw_pile != -1)[0].shape[0]) - 1,
                )
            )
        env[67] = 0
    elif action == 7:
        #  print(f'Phase: {env[67]} Player {env[57]} use two of a kind!')
        env[67] = 2
    elif action == 8:
        #  print(f'Phase: {env[67]} Player {env[57]} use three of a kind!')
        env[67] = 2
    elif action == 9:
        #  print(f'Phase: {env[67]} Player {env[57]} use five different cards!')
        env[67] = 3
    return env, draw_pile, discard_pile


@njit
def idPlayerCanUseNope(env, nope_id):
    """return the id of the player that have the nope card, else -1"""
    main_id = env[57]
    nope_turn = nopeTurn(main_id)
    idx_old = -1
    for i in range(getAgentSize() - 1):
        if nope_turn[i] == nope_id:
            idx_old = i
            break
    else:
        idx_old = -1
    if idx_old + 1 == getAgentSize() - 1:
        return main_id
    else:
        for i in range(idx_old + 1, getAgentSize() - 1):
            idx = nope_turn[i]
            if np.where(env[0:5] == idx)[0].shape[0] >= 1 and env[62:67][int(idx)] == 1:
                return idx
        return main_id


@njit
def discardCardNormalAction(env, last_action, discard_pile):
    if last_action == 0:
        discard_pile[0] += 1
        env[0:5][np.where(env[0:5] == env[57])[0][0]] = 6
    elif last_action == 1:  # Attack
        env[5:9][np.where(env[5:9] == env[57])[0][0]] = 6
        discard_pile[1] += 1
    elif last_action == 2:  # Skip
        env[9:13][np.where(env[9:13] == env[57])[0][0]] = 6
        discard_pile[2] += 1
    elif last_action == 3:
        discard_pile[3] += 1
        env[13:17][np.where(env[13:17] == env[57])[0][0]] = 6
    elif last_action == 4:  # Shuffle
        env[17:21][np.where(env[17:21] == env[57])[0][0]] = 6
        discard_pile[4] += 1
    elif last_action == 5:  # See the future
        env[21:26][np.where(env[21:26] == env[57])[0][0]] = 6
        discard_pile[5] += 1


@njit
def stepEnv(env, draw_pile, discard_pile, action):
    phase = env[67]
    main_id = env[57]
    nope_id = env[73]
    nope_count = env[56]
    last_action = env[72]
    if phase == 0:  # Phase 0: Main Turn
        if action == 6:  # draw card
            env, draw_pile, discard_pile = drawCard(env, draw_pile, discard_pile)
            #  print(f'Phase {phase} Player {main_id} end turn!')
        else:
            env[72] = action
            #  print(f'Phase {phase} Player {main_id} choose action {action}')
            if env[72] >= 7 and env[72] <= 9:
                env[67] = 4  # change to discard phase
                if env[72] == 7:
                    env[75] = 2
                elif env[72] == 8:
                    env[75] = 3
                elif env[72] == 9:
                    env[75] = 5
            elif env[72] >= 0 and env[72] < 7:
                discardCardNormalAction(env, env[72], discard_pile)
                env[76:87] = 0
                env[73] = idPlayerCanUseNope(env, main_id)
                if env[73] == main_id:
                    env, draw_pile, discard_pile = executeMainAction(
                        env, draw_pile, discard_pile, env[72]
                    )
                    #  print(f'Phase: {0} Action {action} has been executed!')
                    #  #  print('ok')
                else:
                    env[67] = 1  # nope phase
    elif phase == 1:  # Phase 1: Nope phase
        if action == 0 and env[73] != main_id:  # other player use Nope
            #  print(f'Phase: {env[67]} Player {env[73]} use Nope!')
            env[56] += 1  #  increase Nope Count
            env[0:5][np.where(env[0:5] == env[73])[0][0]] = 6
            discard_pile[0] += 1
            env[73] = idPlayerCanUseNope(env, env[73])
            if env[73] == main_id:
                if not checkIfNope(env):  # if not been Nope
                    env, draw_pile, discard_pile = executeMainAction(
                        env, draw_pile, discard_pile, env[72]
                    )
                    #  print(f'Phase: {0} Action {env[72]} has been executed!')

        elif action == 0 and env[73] == main_id:
            env[56] += 1  #  increase Nope Count
            env[0:5][np.where(env[0:5] == env[57])[0][0]] = 6
            if not checkIfNope(env):  # if not been Nope
                env, draw_pile, discard_pile = executeMainAction(
                    env, draw_pile, discard_pile, env[72]
                )
                #  print(f'Phase: {0} Action {last_action} has been executed! ')
        elif action == 10:
            if env[73] == main_id:
                if not checkIfNope(env):  # if not been Nope
                    env, draw_pile, discard_pile = executeMainAction(
                        env, draw_pile, discard_pile, env[72]
                    )
                    #  print(f'Phase: {0} Action {last_action} has been executed! ')
                else:  #  if Nope
                    if action == 0:
                        #  print('Main player use Yup!')
                        env[56] = 0  # reset to original
                        env[0:5][np.where(env[0:5] == main_id)[0][0]] = 6
                        discard_pile[0] += 1
                        env[73] = idPlayerCanUseNope(env, env[73])
                        if env[73] == main_id:
                            env, draw_pile, discard_pile = executeMainAction(
                                env, draw_pile, discard_pile, env[72]
                            )
                            #  print(f'Phase: {0} Action {last_action} has been executed! ')
                    else:
                        #  print(f'Phase: {env[67]} Action {last_action} has been Nope!')
                        env[72] = -1  #  action has been Nope
                        env[67] = 0  #  back to phase 0
                        env[73] = idPlayerCanUseNope(env, main_id)
                        env[56] = 0  # reset nope count
            #  env[73] = idPlayerCanUseNope(env,env[73])
            else:
                env[73] = idPlayerCanUseNope(env, env[73])
                if env[73] == main_id:
                    if not checkIfNope(env):  # if not been Nope
                        env, draw_pile, discard_pile = executeMainAction(
                            env, draw_pile, discard_pile, env[72]
                        )

                        #  #  print(f'Phase: {0} Action {env[72]} has been executed! ok3')

    elif (
        phase == 2
    ):  #  phase 2: choose player to steal card. Only main_id can enter this phase
        if action == 6:
            env, draw_pile, discard_pile = drawCard(env, draw_pile, discard_pile)
        else:
            env[74] = env[58:62][int(action - 11)]
            last_action = env[72]
            #  print(f'Phase: {env[67]} Player {env[57]} choose player {env[74]} to steal!')
            if last_action == 7:
                card_on_player_chosen = np.where(env[0:56] == env[74])[0]
                card = np.random.choice(card_on_player_chosen)
                env[0:56][card] = env[57]
                # used card go to Discard Pile
                env[67] = 0
                env[72] = -1
            else:
                env[67] = 3

    elif (
        phase == 3
    ):  # phase 3: choose card to give/take. Only main_id can enter this phase
        last_action = env[72]
        if last_action == 3:
            type_card = action - 15
            low, high = getCardRange(type_card)  # range
            all_card_to_take = np.where(env[low:high] == env[74])[0]
            env[low:high][int(all_card_to_take[0])] = env[57]

        elif last_action == 8:
            # take card
            type_card = action - 27
            low, high = getCardRange(type_card)  # range
            all_card_to_take = np.where(env[low:high] == env[74])[0]
            if all_card_to_take.shape[0] > 0:
                env[low:high][int(all_card_to_take[0])] = env[57]
            # used card go to Discard Pile
        elif last_action == 9:
            type_card = action - 39
            low, high = getCardRange(type_card)  # range
            if np.where(env[low:high] == 6)[0].shape[0] > 0:
                env[low:high][np.where(env[low:high] == 6)[0][0]] = env[57]
                discard_pile[int(type_card)] -= 1
        env[72] = -1
        env[67] = 0

    elif phase == 4:
        all_num_card = getAllNumCard(env, main_id)[:11]
        type_card = action - 51
        #  try:
        #      #  print(f'Phase: {env[67]} Card {visualCard(np.array([low]))} has been discarded!')
        #  except:
        #      #  print(type_card)
        #      raise NameError
        low, high = getCardRange(type_card)  # range
        if last_action == 9:
            env[low:high][np.where(env[low:high] == main_id)[0][0]] = 6
            discard_pile[int(type_card)] += 1
            env[76:87][int(type_card)] = 1
            #  print(env[76:87])
            env[75] -= 1
        else:
            if last_action == 7:
                for i in range(2):
                    env[low:high][np.where(env[low:high] == env[57])[0][0]] = 6
                    discard_pile[int(type_card)] += 1

            elif last_action == 8:
                for i in range(3):
                    env[low:high][np.where(env[low:high] == env[57])[0][0]] = 6
                    discard_pile[int(type_card)] += 1
            env[75] = 0

        if env[75] == 0:
            env[76:87] = 0
            env[73] = idPlayerCanUseNope(env, main_id)
            if env[73] == main_id:
                env, draw_pile, discard_pile = executeMainAction(
                    env, draw_pile, discard_pile, env[72]
                )
                #  print(f'Phase: {0} Action {last_action} has been executed!')
                #  #  print('ok')
            else:
                env[67] = 1  # nope phase
    return env, draw_pile, discard_pile


@njit
def getAgentSize():
    return 5


@njit
def checkEnded(env):
    if np.sum(env[62:67]) == 1:
        return int(np.where(env[62:67] == 1)[0][0])
    else:
        return -1


@njit
def getReward(state):
    if state[86] == 0:
        return 0
    elif np.sum(state[82:86]) == 0:
        return 1
    else:
        return -1


def visualCard(card):
    arr = []
    lst = [
        "Nope",
        "Attack",
        "Skip",
        "Favor",
        "Shuffle",
        "See the future",
        "TCT",
        "RRC",
        "BC",
        "HPC",
        "CTM",
        "Defuse",
        "Exploding Kitten",
    ]
    for i in card:
        if i != -1:
            arr.append(lst[int(getCardType(i))])
    return arr


@njit
def random_player1(state, per):
    list_action = np.where(getValidActions(state) == 1)[0]
    action = np.random.choice(list_action)
    #  print(state[82:86],state[87:91])
    if getReward(state) != -1:
        per += 1
    #  print(getReward(state))
    return action, per


#  @njit
def random_player(state, per):
    list_action = np.where(getValidActions(state) == 1)[0]
    action = np.random.choice(list_action)

    return action, per


@njit()
def one_game_numba(p0, pIdOrder, per_player, per1, per2, per3, per4, p1, p2, p3, p4):
    env, draw_pile, discard_pile = initEnv()

    winner = -1
    turn = 0
    while True:
        turn += 1
        phase = env[67]
        main_id = env[57]
        nope_id = env[73]
        last_action = env[72]
        if phase == 0:
            pIdx = int(main_id)
        elif phase == 1:
            pIdx = int(nope_id)
        elif phase == 2:
            pIdx = int(main_id)
        elif phase == 3:
            if last_action == 3:
                pIdx = int(env[74])
            else:
                pIdx = int(main_id)
        elif phase == 4:
            pIdx = int(main_id)
        if pIdOrder[pIdx] == -1:
            action, per_player = p0(
                getAgentState(env, draw_pile, discard_pile), per_player
            )
        elif pIdOrder[pIdx] == 1:
            action, per1 = p1(getAgentState(env, draw_pile, discard_pile), per1)
        elif pIdOrder[pIdx] == 2:
            action, per2 = p2(getAgentState(env, draw_pile, discard_pile), per2)
        elif pIdOrder[pIdx] == 3:
            action, per3 = p3(getAgentState(env, draw_pile, discard_pile), per3)
        elif pIdOrder[pIdx] == 4:
            action, per4 = p4(getAgentState(env, draw_pile, discard_pile), per4)

        env, draw_pile, discard_pile = stepEnv(env, draw_pile, discard_pile, action)
        winner = checkEnded(env)
        if winner != -1 or turn > 500:
            break
    env[67] = 0
    for pIdx in range(5):
        env[57] = pIdx
        env[58:62] = nopeTurn(env[57])
        if pIdOrder[pIdx] == -1:
            if pIdOrder[pIdx] == -1:
                action, per_player = p0(
                    getAgentState(env, draw_pile, discard_pile), per_player
                )
            elif pIdOrder[pIdx] == 1:
                action, per1 = p1(getAgentState(env, draw_pile, discard_pile), per1)
            elif pIdOrder[pIdx] == 2:
                action, per2 = p2(getAgentState(env, draw_pile, discard_pile), per2)
            elif pIdOrder[pIdx] == 3:
                action, per3 = p3(getAgentState(env, draw_pile, discard_pile), per3)
            elif pIdOrder[pIdx] == 4:
                action, per4 = p4(getAgentState(env, draw_pile, discard_pile), per4)
    win = False
    if np.where(pIdOrder == -1)[0][0] == checkEnded(env):
        win = True
    else:
        win = False
    return win, per_player


@njit()
def n_game_numba(
    p0, num_game, per_player, list_other, per1, per2, per3, per4, p1, p2, p3, p4
):
    win = 0
    for _n in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_numba(
            p0, list_other, per_player, per1, per2, per3, per4, p1, p2, p3, p4
        )
        win += winner
    return win, per_player


import importlib.util
import json, sys

try:
    from env import SHORT_PATH
except:
    pass


def load_module_player(player, game_name=None):
    if game_name == None:
        spec = importlib.util.spec_from_file_location(
            "Agent_player", f"{SHORT_PATH}src/Agent/{player}/Agent_player.py"
        )
    else:
        spec = importlib.util.spec_from_file_location(
            "Agent_player", f"{SHORT_PATH}src/Agent/ifelse/{game_name}/{player}.py"
        )
    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    return module


@njit()
def bot_lv0(state, perData):
    validActions = getValidActions(state)
    arr_action = np.where(validActions == 1)[0]
    idx = np.random.randint(0, arr_action.shape[0])
    return arr_action[idx], perData


@njit()
def check_run_under_njit(Agent):
    return True


def one_game_normal(p0, pIdOrder, per_player, per1, per2, per3, per4, p1, p2, p3, p4):
    env, draw_pile, discard_pile = initEnv()

    winner = -1
    turn = 0
    while True:
        turn += 1
        phase = env[67]
        main_id = env[57]
        nope_id = env[73]
        last_action = env[72]
        if phase == 0:
            pIdx = int(main_id)
        elif phase == 1:
            pIdx = int(nope_id)
        elif phase == 2:
            pIdx = int(main_id)
        elif phase == 3:
            if last_action == 3:
                pIdx = int(env[74])
            else:
                pIdx = int(main_id)
        elif phase == 4:
            pIdx = int(main_id)

        if pIdOrder[pIdx] == -1:
            action, per_player = p0(
                getAgentState(env, draw_pile, discard_pile), per_player
            )
        elif pIdOrder[pIdx] == 1:
            action, per1 = p1(getAgentState(env, draw_pile, discard_pile), per1)
        elif pIdOrder[pIdx] == 2:
            action, per2 = p2(getAgentState(env, draw_pile, discard_pile), per2)
        elif pIdOrder[pIdx] == 3:
            action, per3 = p3(getAgentState(env, draw_pile, discard_pile), per3)
        elif pIdOrder[pIdx] == 4:
            action, per4 = p4(getAgentState(env, draw_pile, discard_pile), per4)
        #  print(f'Nope count: {env[56]}')
        #  print(f'Player {pIdx} Valid Action: {np.where(getValidActions(getAgentState(env,draw_pile,discard_pile))==1)[0]}')
        #  print(f'Turn {turn}, Phase {int(phase)} Player {pIdx} action {action} List player {env[62:67]}' )
        #  print(f'See the future: {visualCard(env[69:72])}')
        #  print()
        env, draw_pile, discard_pile = stepEnv(env, draw_pile, discard_pile, action)
        #  print(f'Draw pile: {visualCard(draw_pile)}')
        #  print("____________________________________________________")
        #  print()

        winner = checkEnded(env)
        if winner != -1 or turn > 500:
            #  print(f'Winner: {winner}')
            break
    env[67] = 0
    for pIdx in range(5):
        env[57] = pIdx
        env[58:62] = nopeTurn(env[57])
        if pIdOrder[pIdx] == -1:
            if pIdOrder[pIdx] == -1:
                action, per_player = p0(
                    getAgentState(env, draw_pile, discard_pile), per_player
                )
            elif pIdOrder[pIdx] == 1:
                action, per1 = p1(getAgentState(env, draw_pile, discard_pile), per1)
            elif pIdOrder[pIdx] == 2:
                action, per2 = p2(getAgentState(env, draw_pile, discard_pile), per2)
            elif pIdOrder[pIdx] == 3:
                action, per3 = p3(getAgentState(env, draw_pile, discard_pile), per3)
            elif pIdOrder[pIdx] == 4:
                action, per4 = p4(getAgentState(env, draw_pile, discard_pile), per4)
    win = False
    if np.where(pIdOrder == -1)[0][0] == checkEnded(env):
        win = True
    else:
        win = False
    return win, per_player


def n_game_normal(
    p0, num_game, per_player, list_other, per1, per2, per3, per4, p1, p2, p3, p4
):
    win = 0
    for _n in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_normal(
            p0, list_other, per_player, per1, per2, per3, per4, p1, p2, p3, p4
        )
        win += winner
    return win, per_player


def load_agent(level, *args):
    num_bot = getAgentSize() - 1
    if "_level_" not in globals():
        global _level_
        _level_ = level
        init = True
    else:
        if _level_ != level:
            _level_ = level
            init = True
        else:
            init = False

    if init:
        global _list_per_level_
        global _list_bot_level_
        _list_per_level_ = []
        _list_bot_level_ = []

        if _level_ == 0:
            _list_per_level_ = [
                np.array([[0.0]], dtype=np.float64) for _ in range(num_bot)
            ]
            _list_bot_level_ = [bot_lv0 for _ in range(num_bot)]
        else:
            env_name = sys.argv[1]
            if len(args) > 0:
                dict_level = json.load(
                    open(f"{SHORT_PATH}src/Log/check_system_about_level.json")
                )
            else:
                dict_level = json.load(open(f"{SHORT_PATH}src/Log/level_game.json"))

            if str(_level_) not in dict_level[env_name]:
                raise Exception("Hiện tại không có level này")

            lst_agent_level = dict_level[env_name][str(level)][2]
            lst_module_level = [
                load_module_player(lst_agent_level[i]) for i in range(num_bot)
            ]
            for i in range(num_bot):
                data_agent_level = np.load(
                    f"{SHORT_PATH}src/Agent/{lst_agent_level[i]}/Data/{env_name}_{level}/Train.npy",
                    allow_pickle=True,
                )
                _list_per_level_.append(
                    lst_module_level[i].convert_to_test(data_agent_level)
                )
                _list_bot_level_.append(lst_module_level[i].Test)

    return _list_bot_level_, _list_per_level_


def run(
    p0: any = bot_lv0,
    num_game: int = 100,
    per_player: np.ndarray = np.array([[0.0]]),
    level: int = 0,
    *args,
):
    num_bot = getAgentSize() - 1
    list_other = np.array([-1] + [i + 1 for i in range(num_bot)])
    try:
        check_njit = check_run_under_njit(p0)
    except:
        check_njit = False

    load_agent(level, *args)

    if check_njit:
        return n_game_numba(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_per_level_[3],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
            _list_bot_level_[3],
        )
    else:
        return n_game_normal(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_per_level_[3],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
            _list_bot_level_[3],
        )



import subprocess
import json
from git import Repo
from tests.CheckEnv import check_pytest
from tests.CheckGraphics import check_graphics
from tests.CheckPlayer import check_agent


def get_changed_files():
    repo_path = ""
    repo = Repo(repo_path)
    latest_commit = repo.head.commit

    commit = f"{latest_commit.hexsha}"
    api_path = f"https://api.github.com/repos/ngoxuanphong/ENV/commits/{commit}"
    command = f"curl -s {api_path}"
    output = subprocess.check_output(
        command, shell=True, stderr=subprocess.STDOUT
    ).decode("utf-8")

    output = json.loads(output)

    print("Last commit: ", latest_commit.hexsha)
    print("API:", api_path)
    changed_files = []
    for file in output["files"]:
        print(file["filename"])
        changed_files.append(file["filename"])

    return changed_files


def test_print_name():
    changed_files = get_changed_files()
    for file in changed_files:
        if "src/Base/" in file and "/env.py" in file:
            env_name = file.replace("src/Base/", "").replace("/env.py", "")
            print(env_name, "checking...")
            check_env, list_bug = check_pytest(env_name)
            if check_env == False:
                print("ENV:", env_name, "FALSE:", list_bug)
                assert False
            else:
                print("ENV:", env_name, "TRUE")
                assert True

        if "src/Agent/" in file and "/Agent_player.py" in file:
            agent_name = file.replace("src/Agent/", "").replace("/Agent_player.py", "")
            print(agent_name, "checking...")
            bool_check_agent, list_bug = check_agent(agent_name)
            if bool_check_agent == False:
                print("Agent:", agent_name, "FALSE:", list_bug)
                assert False
            else:
                print("Agent:", agent_name, "TRUE")
                assert True



import sys
import numpy as np
from numba import njit
import env

game_name = sys.argv[1]
env.make(game_name)


from numba.typed import List


def DataAgent():
    perx_ = List(
        [
            np.random.choice(
                env.getActionSize(), size=env.getActionSize(), replace=False
            ).reshape(1, -1)
            * 1.0,
            np.random.rand(env.getStateSize(), env.getActionSize()) * 2 - 1,
            np.random.rand(env.getActionSize(), env.getActionSize()),
            np.zeros((1, env.getActionSize())),
            np.zeros((1, 1)),
            np.zeros((1, 1)),
            np.zeros((1, 1)),
            np.zeros((1, 1)),
            np.zeros((1, 1)),
            np.random.rand(env.getStateSize(), env.getActionSize()) * 2 - 1,
            np.random.rand(env.getActionSize(), env.getActionSize()),
        ]
    )
    return perx_


@njit
def Train(state, per):
    if per[5][0][0] <= 10000:  # phase1
        actions = env.getValidActions(state)
        if per[4][0][0] == 0:
            action = np.argmax(actions * per[0][0] + actions)
        else:
            output = np.random.rand(env.getActionSize()) * actions + actions
            action = np.argmax(output)
        per[4][0][0] += 1
        if env.getReward(state) != -1:
            per[4][0][0] = 0
            if env.getReward(state) == 1:
                per[3][0] += per[0][0]
            else:
                per[0][0] = (
                    np.random.choice(
                        env.getActionSize(), size=env.getActionSize(), replace=False
                    )
                    * 1.0
                )
            per[5][0][0] += 1
            if per[5][0][0] == 10000:
                per[0][0] = per[3][0] / np.max(per[3][0])
                per[3][0] = per[0][0].copy()

    elif per[5][0][0] > 10000 and per[5][0][0] <= 110000:  # phase 2
        actions = state.reshape(1, -1).dot(per[1])
        per[0][0] += actions[0] / np.max(actions)
        list_action = np.where(env.getValidActions(state) == 1)[0]
        action = list_action[np.argmax(per[0][0][list_action])]
        if env.getReward(state) != -1:
            per[5][0][0] += 1
            per[0][0] = per[3][0].copy()
            if env.getReward(state) == 1:
                per[6][0][0] += 1
            if per[5][0][0] % 1000 == 0:
                if per[6][0][0] / 1000 > per[7][0][0]:
                    per[7][0][0] = per[6][0][0] / 1000
                    per[9] = per[1].copy()
                per[1] = np.random.rand(env.getStateSize(), env.getActionSize()) * 2 - 1
                per[6][0][0] = 0

    elif per[5][0][0] > 110000 and per[5][0][0] <= 210000:  # phase 3
        actions = state.reshape(1, -1).dot(per[1])
        per[0][0] += actions[0] / np.max(actions)
        list_action = np.where(env.getValidActions(state) == 1)[0]
        action = list_action[np.argmax(per[0][0][list_action])]
        per[0][0] += per[2][int(action)]
        if env.getReward(state) != -1:
            per[5][0][0] += 1
            per[0][0] = per[3][0].copy()
            if env.getReward(state) == 1:
                per[6][0][0] += 1
            if per[5][0][0] % 1000 == 0:
                if per[6][0][0] / 1000 > per[8][0][0]:
                    per[8][0][0] = per[6][0][0] / 1000
                    per[10] = per[2].copy()
                per[2] = np.random.rand(env.getActionSize(), env.getActionSize())
                per[6][0][0] = 0

    else:
        actions = state.reshape(1, -1).dot(per[1])
        per[0][0] += actions[0] / np.max(actions)
        list_action = np.where(env.getValidActions(state) == 1)[0]
        action = list_action[np.argmax(per[0][0][list_action])]
        per[0][0] += per[2][int(action)]
        if env.getReward(state) != -1:
            per[5][0][0] += 1
            per[0][0] = per[3][0].copy()
            if env.getReward(state) == 1:
                per[6][0][0] += 1
            if per[5][0][0] % 1000 == 0:
                if (per[5][0][0] % 100000) % 2 == 1:
                    if per[6][0][0] / 1000 > per[8][0][0]:
                        per[8][0][0] = per[6][0][0] / 1000
                        per[10] = per[2].copy()
                    per[2] = np.random.rand(env.getActionSize(), env.getActionSize())
                    per[6][0][0] = 0
                else:
                    if per[6][0][0] / 1000 > per[7][0][0]:
                        per[7][0][0] = per[6][0][0] / 1000
                        per[9] = per[1].copy()
                    per[1] = (
                        np.random.rand(env.getStateSize(), env.getActionSize()) * 2 - 1
                    )
                    per[6][0][0] = 0
    return action, per


perx = [
    np.random.choice(
        env.getActionSize(), size=env.getActionSize(), replace=False
    ).reshape(1, -1)
    * 1.0,
    np.random.rand(env.getStateSize(), env.getActionSize()) * 2 - 1,
    np.random.rand(env.getActionSize(), env.getActionSize()),
    np.zeros((1, env.getActionSize())),
    np.zeros((1, 1)),
    np.zeros((1, 1)),
    np.zeros((1, 1)),
    np.zeros((1, 1)),
    np.zeros((1, 1)),
    np.random.rand(env.getStateSize(), env.getActionSize()) * 2 - 1,
    np.random.rand(env.getActionSize(), env.getActionSize()),
]

# per[0]: arr ban đầu: size = ActionSize, value in range(ActionSize)
# per[1]: arr weight: np.random.rand(StateSize,ActionSize) * 2 -1
# per[2]: arr bias: np.random.rand(ActionSize,ActionSize)
# per[3]: sum all per[0] victory
# per[4]: count turn in one game
# per[5]: count game
#  per[6]: count win in 1000 games
# per[7]: max win in 1000 games phase 2
# per[8]: max win in 1000 games phase 3
# per[9]: per[1] best.
# per[10]: per[2] best


@njit
def Test(state, per):
    actions = state.reshape(1, -1).dot(per[1])
    per[0][0] += actions[0] / np.max(actions)
    list_action = np.where(env.getValidActions(state) == 1)[0]
    action = list_action[np.argmax(per[0][0][list_action])]
    per[0][0] += per[2][action]
    if env.getReward(state) != -1:
        per[0][0] = per[3][0]
    return action, per


def convert_to_save(perData):
    if len(perData) == 4:
        raise Exception("Data này đã được convert rồi.")
    data = List()
    data.append(perData[0])
    data.append(perData[9])
    data.append(perData[10])
    data.append(perData[3])
    return data


def convert_to_test(perData):
    return List(perData)



import sys
import numpy as np
from numba import njit
import env

game_name = sys.argv[1]
env.make(game_name)


from numba.typed import List


def convert_to_save(perData):
    if len(perData) == 2:
        raise Exception("Data này đã được convert rồi.")

    data = List()
    data.append(np.zeros((1, 1), np.int64))
    temp = perData[2].astype(np.int64)
    k = temp.shape[0]
    for i in range(temp.shape[0]):
        if (temp[i] == 0).all():
            k = i
            break

    temp = temp[0:k]
    data.append(temp)
    return data


def convert_to_test(perData):
    return List(perData)


def DataAgent():
    per = List(
        [
            np.zeros((1, 1)),  # [0][0][0] turn hiện tại
            np.zeros(
                (10000, env.getActionSize())
            ),  # [1][turn] lưu bộ đã dùng trong game
            np.zeros((10000, env.getActionSize())),  # [2][turn] lưu kết quả khi thắng
        ]
    )
    return per


@njit()
def Train(state, per):
    actions = env.getValidActions(state)
    turn = int(per[0][0][0])
    if np.sum(per[1][turn]) == 0:
        weights = np.argsort(np.argsort(np.random.rand(env.getActionSize()))) * 1.0
        per[1][turn] = weights
    else:
        weights = per[1][turn]
    output = actions * weights + actions
    action = np.argmax(output)
    win = env.getReward(state)
    per[0] += 1
    if win != -1:
        if win == 1:
            per[2] += per[1]
        if win == 0:
            per[1] = np.zeros((10000, env.getActionSize()))
        per[0] = np.zeros((1, 1))
    return action, per


@njit()
def Test(state, per):
    actions = env.getValidActions(state)
    turn = per[0][0][0]
    if per[1].shape[0] > turn:
        weights = per[1][turn]
        output = actions * weights + actions
        action = np.argmax(output)
    else:
        actions = np.where(actions == 1)[0]
        action = actions[np.random.randint(0, actions.shape[0])]

    win = env.getReward(state)
    per[0][0][0] += 1
    if win != -1:
        per[0][0][0] = 0

    return action, per



import numba as nb
import numpy as np
from numba import njit

from src.Base.Century.docs.index import *


@njit()
def getAgentSize():
    return 5


@njit()
def getStateSize():
    return 277


@njit()
def getActionSize():
    return 65


# khởi tạo bàn chơi
@njit()
def initEnv():
    start_card_player = np.concatenate((np.array([1]), np.zeros(42), np.array([1, 0])))
    start_player_0 = np.concatenate((np.array([0, 0, 3, 0, 0, 0]), start_card_player))
    start_player_1 = np.concatenate((np.array([0, 0, 4, 0, 0, 0]), start_card_player))
    start_player_2 = np.concatenate((np.array([0, 0, 4, 0, 0, 0]), start_card_player))
    start_player_3 = np.concatenate((np.array([0, 0, 3, 1, 0, 0]), start_card_player))
    start_player_4 = np.concatenate((np.array([0, 0, 3, 1, 0, 0]), start_card_player))
    list_card = np.append(np.arange(1, 43), 44)
    list_card_point = np.arange(36)
    np.random.shuffle(list_card)
    np.random.shuffle(list_card_point)
    top_6_card = ALL_CARD_IN4[list_card[:NUMBER_OPEN_ACTION_CARD]].flatten()
    top_5_card_point = ALL_CARD_POINT_IN4[
        list_card_point[:NUMBER_OPEN_POINT_CARD]
    ].flatten()
    order_player = np.arange(NUMBER_PLAYER)
    env_state = np.concatenate(
        (
            start_player_0,
            start_player_1,
            start_player_2,
            start_player_3,
            start_player_4,
            top_6_card,
            np.zeros(20),
            top_5_card_point,
            list_card,
            list_card_point,
            np.array([0, -0.5, 0, 10, 10, -1, 1, 0, 0]),
            order_player,
        )
    )
    # 5 player_in4, 6card_in4,5 token free, 5 card_point_in4, list_card_shuffle, list_card_point_shuffle, [number_action_upgrade, card_will_buy/card_hand_used, token need drop, silver, gold, last_action, phase, check_end, id_action]
    return env_state


@njit()
def getAgentState(env_state):
    player_state = np.zeros(P_LENGTH, dtype=np.float64)
    player_action = int(env_state[ENV_ID_ACTION])  # xác định người chơi
    player_main_in4 = env_state[
        ATTRIBUTE_PLAYER * player_action : ATTRIBUTE_PLAYER * (player_action + 1)
    ]
    player_state[P_SCORE : P_SCORE + BASIC_ATTRIBUTE_PLAYER] = player_main_in4[
        :BASIC_ATTRIBUTE_PLAYER
    ]
    player_state[P_ACTION_CARD_PLAYER:P_ACTION_CARD_DOWN_PLAYER] = (
        player_main_in4[BASIC_ATTRIBUTE_PLAYER:] > 0
    ) * 1
    player_state[P_ACTION_CARD_DOWN_PLAYER:P_OTHER_PLAYER_IN4] = (
        player_main_in4[BASIC_ATTRIBUTE_PLAYER:] < 0
    ) * 1
    for idx in range(1, 5):
        id = int((player_action + idx) % NUMBER_PLAYER)
        all_other_player_in4 = env_state[
            ATTRIBUTE_PLAYER * id : ATTRIBUTE_PLAYER * id + BASIC_ATTRIBUTE_PLAYER
        ].copy()
        # all_other_player_card = np.where(all_other_player_in4[BASIC_ATTRIBUTE_PLAYER:] == -1, 1, 0)      #xác định thẻ người chơi có dựa trên các thẻ đã đánh
        player_state[
            P_OTHER_PLAYER_IN4
            + (idx - 1) * BASIC_ATTRIBUTE_PLAYER : P_OTHER_PLAYER_IN4
            + idx * BASIC_ATTRIBUTE_PLAYER
        ] = all_other_player_in4
    # player_state = np.concatenate((player_state, env_state[ENV_OPEN_ACTION_CARD:ENV_DOWN_ACTION_CARD], env_state[ENV_SILVER_COIN : ENV_ID_ACTION]))
    player_state[P_OPEN_ACTION_CARD:P_SILVER_COIN] = env_state[
        ENV_OPEN_ACTION_CARD:ENV_DOWN_ACTION_CARD
    ]
    player_state[P_SILVER_COIN] = env_state[ENV_SILVER_COIN]
    player_state[P_GOLD_COIN] = env_state[ENV_GOLD_COIN]
    # action gần nhất nếu dùng thẻ action đổi token
    if env_state[ENV_LAST_ACTION] != -1:
        player_state[
            P_LAST_ACTION + int(env_state[ENV_LAST_ACTION])
        ] = 1  # update 25/4/2023
    # kiểm tra end game
    player_state[P_CHECK_END] = env_state[ENV_CHECK_END]

    # THứ tự người chơi
    player_state[P_ORDER : P_ORDER + NUMBER_PLAYER] = np.concatenate(
        (
            env_state[
                ENV_ORDER_PLAYER + player_action : ENV_ORDER_PLAYER + NUMBER_PLAYER
            ],
            env_state[ENV_ORDER_PLAYER : ENV_ORDER_PLAYER + player_action],
        )
    )
    # phase
    player_state[P_PHASE + int(env_state[ENV_PHASE]) - 1] = 1
    # print(list(player_state))
    # print(np.sum(player_state[221:266]), player_state[221:266])
    return player_state


@njit()
def getValidActions(player_state_origin):
    list_action_return = np.zeros(65, dtype=np.float64)
    player_state = player_state_origin.copy()
    phase_env = -1
    if len(np.where(player_state[P_PHASE : P_PHASE + NUMBER_PHASE] == 1)[0]) != 0:
        phase_env = (
            np.where(player_state[P_PHASE : P_PHASE + NUMBER_PHASE] == 1)[0][0] + 1
        )
    player_state_own = player_state[:P_OTHER_PLAYER_IN4]
    """
        Quy ước phase: 
        phase1: chọn mua thẻ (6 thẻ top và 5 thẻ point) hay đánh thẻ (thẻ trên tay) hay nghỉ ngơi (11 action mua, 45 action đánh, 1 action nghỉ)
        phase2: nếu mua thẻ top, chọn token để vào các thẻ trước thẻ mình mua (4 action)
        phase3: nếu đánh thẻ, chọn xem có thực hiện action của thẻ tiếp ko (2action, 1 cái là ko, 1 cái trùng vs action dùng thẻ)
        phase4: trả token dư thừa (4 action) (trùng phase 2)
        phase5: chọn tài nguyên nâng cấp    
    """
    player_token = player_state_own[P_TOKEN:P_ACTION_CARD_PLAYER]
    if phase_env == 1:
        # chọn mua thẻ (6 thẻ top và 5 thẻ point) hay đánh thẻ (thẻ trên tay) hay nghỉ ngơi (11 action mua, 45 action đánh, 1 action nghỉ)
        list_action_return[0] = 1  # luôn có action nghỉ ngơi
        # check mua 6 thẻ top
        number_token = np.sum(player_token)
        card_on_board = player_state[P_OPEN_ACTION_CARD:P_TOKEN_ON_ACTION_CARD]
        for act in range(6):
            if (
                act <= number_token
                and np.sum(
                    card_on_board[
                        LENGTH_ACTION_CARD * act : LENGTH_ACTION_CARD * (act + 1)
                    ]
                )
                > 0
            ):  # kiểm tra đủ tài nguyên đặt không, bàn chơi có còn thẻ ko
                list_action_return[act + FIRST_ACTION_BUY_ACTION_CARD] = 1
            else:
                break
        # check mua 5 thẻ point
        all_card_point = player_state[
            P_OPEN_POINT_CARD : P_OPEN_POINT_CARD
            + NUMBER_OPEN_POINT_CARD * ATTRIBUTE_POINT_CARD
        ]
        for id in range(5):
            card_in4 = all_card_point[
                ATTRIBUTE_POINT_CARD * id : ATTRIBUTE_POINT_CARD * (id + 1)
            ][:NUMBBER_TYPE_TOKEN]
            if np.sum(card_in4 > player_token) == 0:
                list_action_return[id + FIRST_ACTION_BUY_POINT_CARD] = 1
        # check đánh thẻ trên tay
        data = ALL_CARD_IN4.copy()
        for card_hand in range(NUMBER_ACTION_CARD - NUMBER_UPGRADE_CARD):
            if player_state[P_ACTION_CARD_PLAYER + card_hand] == 1:
                give = data[card_hand][:NUMBBER_TYPE_TOKEN]
                if np.sum(give > player_token) == 0:
                    list_action_return[card_hand + FIRST_ACTION_USE_ACTION_CARD] = 1

        if player_state[ATTRIBUTE_PLAYER - NUMBER_UPGRADE_CARD] == 1:
            if np.sum(player_token[:3] > 0) != 0:
                list_action_return[55] = 1
        if player_state[ATTRIBUTE_PLAYER - NUMBER_UPGRADE_CARD + 1] == 1:
            if np.sum(player_token[:3] > 0) != 0:
                list_action_return[56] = 1

    elif phase_env == 2:
        # nếu mua thẻ top và cần bỏ token, chọn token để vào các thẻ trước thẻ mình mua (4 action)
        list_action = np.where(player_token > 0)[0] + FIRST_ACTION_DROP_TOKEN
        list_action_return[list_action] = 1

    elif phase_env == 3:
        # nếu đánh thẻ, chọn xem có thực hiện action của thẻ tiếp ko (2action, 1 cái là ko, 1 cái trùng vs action dùng thẻ)
        list_action_return[61] = 1
        list_action_return[
            FIRST_ACTION_USE_ACTION_CARD : FIRST_ACTION_USE_ACTION_CARD
            + NUMBER_ACTION_CARD
        ] = (player_state[P_LAST_ACTION : P_LAST_ACTION + NUMBER_ACTION_CARD] == 1) * 1

    elif phase_env == 4:
        # trả token dư thừa (4 action) sau khi đánh thẻ hoặc mua thẻ top
        list_action = np.where(player_token > 0)[0] + FIRST_ACTION_DROP_TOKEN
        list_action_return[list_action] = 1

    elif phase_env == 5:
        # chọn token để nâng cấp
        list_action = np.where(player_token[:3] > 0)[0] + FIRST_ACTION_UPGRADE_TOKEN
        list_action_return[list_action] = 1

    return list_action_return


@njit()
def getReward(player_state):
    value_return = -1
    end_true = 0
    # kiểm tra xem game kết thúc thật hay loop quá giới hạn
    if player_state[P_PLAYER_NUMBER_POINT_CARD] == 5:
        end_true = 1
    else:
        for id in range(NUMBER_PLAYER - 1):
            if (
                player_state[P_OTHER_PLAYER_IN4:][
                    id * BASIC_ATTRIBUTE_PLAYER : (id + 1) * BASIC_ATTRIBUTE_PLAYER
                ][1]
                == 5
            ):
                end_true = 1
                break

    if player_state[P_CHECK_END] != 1:
        return value_return
    else:
        if end_true == 0:  # nếu loop quá lâu thì auto thua
            return 0
        else:
            # all_player_point = np.zeros(NUMBER_PLAYER)
            all_player_point = player_state[P_ORDER : P_ORDER + NUMBER_PLAYER].copy()
            for id_player in range(5):
                player_in4 = np.zeros(BASIC_ATTRIBUTE_PLAYER)
                if id_player == 0:
                    player_in4 = player_state[:BASIC_ATTRIBUTE_PLAYER]
                else:
                    player_in4 = player_state[
                        P_OTHER_PLAYER_IN4
                        + (id_player - 1) * BASIC_ATTRIBUTE_PLAYER : P_OTHER_PLAYER_IN4
                        + id_player * BASIC_ATTRIBUTE_PLAYER
                    ]

                player_point = player_in4[0] + np.sum(player_in4[3:6]) * 10
                all_player_point[id_player] += player_point
            if np.argmax(all_player_point) == 0:
                return 1
            else:
                return 0


@njit()
def check_winner(env_state):
    winner = -1
    end = 0
    for id_player in range(5):
        player_number_card_point = env_state[ATTRIBUTE_PLAYER * id_player + 1]
        if player_number_card_point == 5:
            end = 1
            break
    if end == 1:
        # all_player_point = np.zeros(NUMBER_PLAYER)
        all_player_point = env_state[
            ENV_ORDER_PLAYER : ENV_ORDER_PLAYER + NUMBER_PLAYER
        ].copy()
        for id_player in range(5):
            player_in4 = env_state[
                ATTRIBUTE_PLAYER * id_player : ATTRIBUTE_PLAYER * (id_player + 1)
            ]
            player_point = player_in4[0] + np.sum(player_in4[3:6]) * 10
            # all_player_point[id_player] = player_point
            all_player_point[id_player] += player_point
        winner = np.argmax(all_player_point)
        return winner
    else:
        # print('một ván ko kết thúc')
        return winner


@njit()
def system_check_end(env_state):
    for id_player in range(5):
        if (
            env_state[
                ATTRIBUTE_PLAYER * id_player : ATTRIBUTE_PLAYER * (id_player + 1)
            ][1]
            == 5
            and env_state[ENV_ID_ACTION] == 0
        ):
            return False
    return True


@njit()
def stepEnv(env_state, action):
    phase_env = int(env_state[ENV_PHASE])
    id_action = int(env_state[ENV_ID_ACTION])
    player_in4 = env_state[
        ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
    ]
    if phase_env == 1:
        # nếu người chơi nghỉ
        if action == 0:
            card_hand_player = player_in4[BASIC_ATTRIBUTE_PLAYER:ATTRIBUTE_PLAYER]
            card_hand_player = np.where(card_hand_player == -1, 1, card_hand_player)
            player_in4[BASIC_ATTRIBUTE_PLAYER:ATTRIBUTE_PLAYER] = card_hand_player
            env_state[
                ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
            ] = player_in4
            env_state[ENV_PHASE] = 1
            env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % 5
        # nếu người chơi mua thẻ trên bàn
        elif action in range(1, 7):
            if action == 1:  # nếu mua thẻ đầu
                # lấy thông tin
                list_card_player = player_in4[BASIC_ATTRIBUTE_PLAYER:ATTRIBUTE_PLAYER]
                list_card_board = env_state[ENV_DOWN_ACTION_CARD:ENV_DOWN_POINT_CARD]
                idx_card_buy = action - 1
                card_buy = int(list_card_board[idx_card_buy])
                all_token_free = env_state[ENV_TOKEN_ON_ACTION_CARD:ENV_OPEN_POINT_CARD]
                token_free = all_token_free[
                    NUMBBER_TYPE_TOKEN
                    * idx_card_buy : NUMBBER_TYPE_TOKEN
                    * (idx_card_buy + 1)
                ]
                all_token_free = np.concatenate(
                    (all_token_free[NUMBBER_TYPE_TOKEN:], np.zeros(4))
                )
                # cập nhật giá trị
                list_card_player[card_buy] = 1
                list_card_board[idx_card_buy:] = np.append(
                    list_card_board[idx_card_buy + 1 :], -1
                )
                top_6_card = np.zeros((NUMBER_OPEN_ACTION_CARD, LENGTH_ACTION_CARD))
                for i in range(6):
                    id = list_card_board[:6][i]
                    top_6_card[i] = ALL_CARD_IN4[int(id)]
                top_6_card = top_6_card.flatten()
                player_in4[
                    BASIC_ATTRIBUTE_PLAYER:
                ] = list_card_player  # cập nhật thẻ mới mua
                player_in4[2:6] += token_free  # cập nhật token free nếu có
                env_state[
                    ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
                ] = player_in4
                env_state[
                    ENV_DOWN_ACTION_CARD:ENV_DOWN_POINT_CARD
                ] = list_card_board  # cập nhật danh sách thẻ trên bàn
                env_state[
                    ENV_OPEN_ACTION_CARD:ENV_TOKEN_ON_ACTION_CARD
                ] = top_6_card  # cập nhật 6 thẻ người chơi có thể mua
                env_state[
                    ENV_TOKEN_ON_ACTION_CARD:ENV_OPEN_POINT_CARD
                ] = all_token_free  # cập nhật giảm token free
                # kiểm tra có phải trả tài nguyên ko
                if np.sum(player_in4[2:BASIC_ATTRIBUTE_PLAYER]) > 10:
                    env_state[ENV_TOKEN_NEED_DROP] = (
                        np.sum(player_in4[2:BASIC_ATTRIBUTE_PLAYER]) - 10
                    )
                    env_state[ENV_PHASE] = 4
                else:
                    # chuyển người chơi
                    env_state[ENV_PHASE] = 1
                    env_state[ENV_ID_ACTION] = (
                        env_state[ENV_ID_ACTION] + 1
                    ) % NUMBER_PLAYER

            else:  # nếu mua thẻ cần đặt token
                # lấy thông tin
                idx_card_buy = action - 1
                # đẩy thông tin vào hệ thống, thông tin là index thẻ mua và số token cần đặt
                env_state[ENV_TOKEN_NEED_DROP] = idx_card_buy
                env_state[ENV_CARD_BUY_OR_USE] = idx_card_buy
                # chuyển phase
                env_state[ENV_PHASE] = 2

        elif action in range(7, 12):
            # lấy thông tin
            list_card_point_board = env_state[
                ENV_DOWN_POINT_CARD : ENV_DOWN_POINT_CARD + NUMBER_POINT_CARD
            ]
            idx_card_buy = action - 7
            card_buy = int(list_card_point_board[idx_card_buy])
            token_fee = ALL_CARD_POINT_IN4[card_buy][:4]
            free_score = 0
            if idx_card_buy < 2:
                if idx_card_buy == 0:
                    if env_state[ENV_GOLD_COIN] != 0:
                        free_score = FREE_SCORE_GOLD
                        env_state[ENV_GOLD_COIN] -= 1
                    else:
                        if env_state[ENV_SILVER_COIN] != 0:
                            free_score = FREE_SCORE_SILVER
                            env_state[ENV_SILVER_COIN] -= 1
                else:
                    if (
                        env_state[ENV_SILVER_COIN] != 0
                        and env_state[ENV_GOLD_COIN] != 0
                    ):
                        free_score = FREE_SCORE_SILVER
                        env_state[ENV_SILVER_COIN] -= 1
            # Cập nhật giá trị
            player_in4[0] += free_score + ALL_CARD_POINT_IN4[card_buy][-1]
            player_in4[1] += 1
            player_in4[2:BASIC_ATTRIBUTE_PLAYER] -= token_fee
            list_card_point_board[idx_card_buy:] = np.append(
                list_card_point_board[idx_card_buy + 1 :], -1
            )
            top_5_card_point = np.zeros((NUMBER_OPEN_POINT_CARD, ATTRIBUTE_POINT_CARD))
            for i in range(NUMBER_OPEN_POINT_CARD):
                id = list_card_point_board[:5][i]
                top_5_card_point[i] = ALL_CARD_POINT_IN4[int(id)]
            top_5_card_point = top_5_card_point.flatten()

            env_state[
                ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
            ] = player_in4
            env_state[
                ENV_DOWN_POINT_CARD : ENV_DOWN_POINT_CARD + NUMBER_POINT_CARD
            ] = list_card_point_board
            env_state[
                ENV_OPEN_POINT_CARD : ENV_OPEN_POINT_CARD
                + NUMBER_OPEN_POINT_CARD * ATTRIBUTE_POINT_CARD
            ] = top_5_card_point
            # Chuyển người chơi
            env_state[ENV_PHASE] = 1
            env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % NUMBER_PLAYER

        elif action in range(12, 57):
            # lấy thông tin
            card_hand_player = player_in4[BASIC_ATTRIBUTE_PLAYER:ATTRIBUTE_PLAYER]
            id_card_use = action - 12
            token_fee_get = ALL_CARD_IN4[id_card_use][:ATTRIBUTE_ACTION_CARD]
            card_hand_player[id_card_use] = -1

            if np.sum(token_fee_get) == 0:  # nếu là thẻ nâng cấp
                player_in4[BASIC_ATTRIBUTE_PLAYER:ATTRIBUTE_PLAYER] = card_hand_player
                env_state[
                    ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
                ] = player_in4
                env_state[ENV_NUMBER_ACTION_UPGRADE] = ALL_CARD_IN4[id_card_use][-1]
                env_state[ENV_PHASE] = 5
            else:
                # Cập nhật giá trị
                player_in4[2:BASIC_ATTRIBUTE_PLAYER] = (
                    player_in4[2:BASIC_ATTRIBUTE_PLAYER]
                    - token_fee_get[:4]
                    + token_fee_get[4:]
                )
                player_in4[6:51] = card_hand_player
                env_state[
                    ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
                ] = player_in4
                # nếu thẻ được dùng nhiều lần
                if (
                    np.sum(token_fee_get[:4]) > 0
                    and np.sum(token_fee_get[:4] > player_in4[2:6]) == 0
                ):
                    env_state[
                        ENV_CARD_BUY_OR_USE
                    ] = id_card_use  # lưu trữ thẻ dùng gần nhất
                    env_state[
                        ENV_LAST_ACTION
                    ] = id_card_use  # lưu trữ action_main gần nhất
                    env_state[ENV_PHASE] = 3  # chuyển phase
                else:  # dùng 1 lần rồi bỏ
                    if (
                        np.sum(player_in4[2:BASIC_ATTRIBUTE_PLAYER]) > 10
                    ):  # nếu thừa nguyên liệu thì đi lược bỏ
                        env_state[ENV_TOKEN_NEED_DROP] = (
                            np.sum(player_in4[2:BASIC_ATTRIBUTE_PLAYER]) - 10
                        )
                        env_state[ENV_PHASE] = 4
                    else:
                        env_state[ENV_PHASE] = 1
                        env_state[ENV_ID_ACTION] = (
                            env_state[ENV_ID_ACTION] + 1
                        ) % NUMBER_PLAYER

    elif phase_env == 2:
        # lấy thông tin
        stay_drop = int(env_state[ENV_TOKEN_NEED_DROP]) - 1
        all_token_free = env_state[ENV_TOKEN_ON_ACTION_CARD:ENV_OPEN_POINT_CARD]
        token_drop = action - 57
        # Cập nhật thông tin
        player_in4[2:BASIC_ATTRIBUTE_PLAYER][token_drop] -= 1
        all_token_free[4 * stay_drop + token_drop] += 1
        env_state[ENV_TOKEN_NEED_DROP] -= 1

        if env_state[ENV_TOKEN_NEED_DROP] == 0:  # Hoàn tất đặt nguyên liệu thì lấy thẻ
            # lấy thông tin
            list_card_player = player_in4[BASIC_ATTRIBUTE_PLAYER:]
            list_card_board = env_state[ENV_DOWN_ACTION_CARD:ENV_DOWN_POINT_CARD]
            idx_card_buy = int(env_state[ENV_CARD_BUY_OR_USE])
            card_buy = int(list_card_board[idx_card_buy])
            token_free = np.zeros(NUMBBER_TYPE_TOKEN)
            if idx_card_buy != 5:
                token_free = all_token_free[
                    NUMBBER_TYPE_TOKEN
                    * idx_card_buy : NUMBBER_TYPE_TOKEN
                    * (idx_card_buy + 1)
                ]
                all_token_free = np.concatenate(
                    (
                        all_token_free[: NUMBBER_TYPE_TOKEN * idx_card_buy],
                        all_token_free[NUMBBER_TYPE_TOKEN * (idx_card_buy + 1) :],
                        np.zeros(NUMBBER_TYPE_TOKEN),
                    )
                )  # cập nhật giảm token free
            # cập nhật giá trị
            list_card_player[card_buy] = 1
            list_card_board[idx_card_buy:] = np.append(
                list_card_board[idx_card_buy + 1 :], -1
            )
            # top_6_card = card_in4[np.array(list_card_board[:6], dtype = int)].flatten()
            top_6_card = np.zeros((NUMBER_OPEN_ACTION_CARD, LENGTH_ACTION_CARD))
            for i in range(6):
                id = list_card_board[:6][i]
                top_6_card[i] = ALL_CARD_IN4[int(id)]

            # top_6_card = np.array([card_in4[int(id)] for id in list_card_board[:6]]).flatten()
            top_6_card = top_6_card.flatten()
            player_in4[
                BASIC_ATTRIBUTE_PLAYER:
            ] = list_card_player  # cập nhật thẻ mới mua
            player_in4[2:BASIC_ATTRIBUTE_PLAYER] = (
                player_in4[2:BASIC_ATTRIBUTE_PLAYER] + token_free
            )  # cập nhật token free nếu có
            env_state[
                ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
            ] = player_in4
            env_state[
                ENV_TOKEN_ON_ACTION_CARD:ENV_OPEN_POINT_CARD
            ] = all_token_free  # cập nhật token free
            env_state[
                ENV_DOWN_ACTION_CARD:ENV_DOWN_POINT_CARD
            ] = list_card_board  # cập nhật danh sách thẻ trên bàn
            env_state[
                ENV_OPEN_ACTION_CARD:ENV_TOKEN_ON_ACTION_CARD
            ] = top_6_card  # cập nhật 6 thẻ người chơi có thể mua
            # Khôi phục các giá trị lưu trữ
            env_state[ENV_CARD_BUY_OR_USE] = -0.5
            # kiểm tra có phải trả tài nguyên ko
            if np.sum(player_in4[2:BASIC_ATTRIBUTE_PLAYER]) > 10:
                env_state[ENV_TOKEN_NEED_DROP] = (
                    np.sum(player_in4[2:BASIC_ATTRIBUTE_PLAYER]) - 10
                )
                env_state[ENV_PHASE] = 4
            else:
                # chuyển người chơi
                env_state[ENV_PHASE] = 1
                env_state[ENV_ID_ACTION] = (
                    env_state[ENV_ID_ACTION] + 1
                ) % NUMBER_PLAYER
        else:
            env_state[
                ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
            ] = player_in4
            env_state[
                ENV_TOKEN_ON_ACTION_CARD:ENV_OPEN_POINT_CARD
            ] = all_token_free  # cập nhật token free

    elif phase_env == 3:
        if action == 61:  # nếu ko action tiếp
            env_state[ENV_CARD_BUY_OR_USE] = -0.5  # lưu trữ thẻ dùng gần nhất
            env_state[ENV_LAST_ACTION] = -1  # lưu trữ action_main gần nhất
            if (
                np.sum(player_in4[2:BASIC_ATTRIBUTE_PLAYER]) > 10
            ):  # nếu thừa nguyên liệu thì đi lược bỏ
                env_state[ENV_TOKEN_NEED_DROP] = (
                    np.sum(player_in4[2:BASIC_ATTRIBUTE_PLAYER]) - 10
                )
                env_state[ENV_PHASE] = 4
            else:
                env_state[ENV_PHASE] = 1
                env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % 5
        else:
            # Lấy thông tin
            id_card_use = int(env_state[ENV_CARD_BUY_OR_USE])
            token_fee_get = ALL_CARD_IN4[id_card_use]
            # Cập nhật thông tin
            player_in4[2:BASIC_ATTRIBUTE_PLAYER] = (
                player_in4[2:BASIC_ATTRIBUTE_PLAYER]
                - token_fee_get[:NUMBBER_TYPE_TOKEN]
                + token_fee_get[NUMBBER_TYPE_TOKEN:ATTRIBUTE_ACTION_CARD]
            )
            env_state[
                ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
            ] = player_in4
            if (
                np.sum(
                    token_fee_get[:NUMBBER_TYPE_TOKEN]
                    > player_in4[2:BASIC_ATTRIBUTE_PLAYER]
                )
                == 0
            ):
                # env_state[-7] = id_card_use     #lưu trữ thẻ dùng gần nhất
                # env_state[-3] = action          #lưu trữ action_main gần nhất
                env_state[ENV_PHASE] = 3  # chuyển phase
            else:
                env_state[ENV_CARD_BUY_OR_USE] = -0.5
                env_state[ENV_LAST_ACTION] = -1
                if np.sum(player_in4[2:6]) > 10:  # nếu thừa nguyên liệu thì đi lược bỏ
                    env_state[ENV_TOKEN_NEED_DROP] = np.sum(player_in4[2:6]) - 10
                    env_state[ENV_PHASE] = 4
                else:
                    env_state[ENV_PHASE] = 1
                    env_state[ENV_ID_ACTION] = (
                        env_state[ENV_ID_ACTION] + 1
                    ) % NUMBER_PLAYER

    elif phase_env == 4:
        # lấy thông tin
        # stay_drop = env_state[-6]
        token_drop = action - 57
        # Cập nhật thông tin
        player_in4[2:6][token_drop] -= 1
        env_state[
            ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
        ] = player_in4
        env_state[ENV_TOKEN_NEED_DROP] -= 1
        if env_state[ENV_TOKEN_NEED_DROP] == 0:
            env_state[ENV_PHASE] = 1
            env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % NUMBER_PLAYER
        else:
            return env_state

    elif phase_env == 5:
        id_update = action - 62
        # if id_update == 3:
        #     env_state[-8] = 0
        #     env_state[-2] = 1
        #     env_state[-1] = (env_state[-1] + 1)%5
        # else:
        player_in4[2:6][id_update] -= 1
        player_in4[2:6][id_update + 1] += 1
        env_state[
            ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
        ] = player_in4
        env_state[ENV_NUMBER_ACTION_UPGRADE] -= 1
        if (
            env_state[ENV_NUMBER_ACTION_UPGRADE] == 0
            or np.sum(player_in4[2:5] > 0) == 0
        ):
            env_state[ENV_PHASE] = 1
            env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % NUMBER_PLAYER

    return env_state


@njit()
def bot_lv0(state, perData):
    validActions = getValidActions(state)
    arr_action = np.where(validActions == 1)[0]
    idx = np.random.randint(0, arr_action.shape[0])
    return arr_action[idx], perData


@njit()
def one_game_numba(p0, list_other, per_player, per1, per2, per3, per4, p1, p2, p3, p4):
    env_state = initEnv()
    count_turn = 0
    while system_check_end(env_state) and count_turn < 2000:
        idx = int(env_state[ENV_ID_ACTION])
        player_state = getAgentState(env_state)
        if list_other[idx] == -1:
            action, per_player = p0(player_state, per_player)
            if getValidActions(player_state)[action] != 1:
                raise Exception("bot dua ra action khong hop le")

        elif list_other[idx] == 1:
            action, per1 = p1(player_state, per1)
        elif list_other[idx] == 2:
            action, per2 = p2(player_state, per2)
        elif list_other[idx] == 3:
            action, per3 = p3(player_state, per3)
        elif list_other[idx] == 4:
            action, per4 = p4(player_state, per4)
        else:
            raise Exception("Sai list_other.")

        env_state = stepEnv(env_state, action)
        count_turn += 1

    env_state[ENV_CHECK_END] = 1

    for p_idx in range(5):
        env_state[ENV_PHASE] = 1
        idx = int(env_state[ENV_ID_ACTION])
        player_state = getAgentState(env_state)
        if list_other[idx] == -1:
            action, per_player = p0(player_state, per_player)
        elif list_other[idx] == 1:
            action, per1 = p1(player_state, per1)
        elif list_other[idx] == 2:
            action, per2 = p2(player_state, per2)
        elif list_other[idx] == 3:
            action, per3 = p3(player_state, per3)
        elif list_other[idx] == 4:
            action, per4 = p4(player_state, per4)
        else:
            raise Exception("Sai list_other.")
        env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % NUMBER_PLAYER

    winner = False
    if np.where(list_other == -1)[0] == check_winner(env_state):
        winner = True
    else:
        winner = False

    return winner, per_player


@njit()
def n_games_numba(
    p0, num_game, per_player, list_other, per1, per2, per3, per4, p1, p2, p3, p4
):
    win = 0
    for _n in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_numba(
            p0, list_other, per_player, per1, per2, per3, per4, p1, p2, p3, p4
        )
        win += winner
    return win, per_player


def one_game_normal(p0, list_other, per_player, per1, per2, per3, per4, p1, p2, p3, p4):
    env_state = initEnv()
    count_turn = 0
    while system_check_end(env_state) and count_turn < 2000:
        idx = int(env_state[ENV_ID_ACTION])
        player_state = getAgentState(env_state)
        if list_other[idx] == -1:
            action, per_player = p0(player_state, per_player)
            if getValidActions(player_state)[action] != 1:
                raise Exception("bot dua ra action khong hop le")

        elif list_other[idx] == 1:
            action, per1 = p1(player_state, per1)
        elif list_other[idx] == 2:
            action, per2 = p2(player_state, per2)
        elif list_other[idx] == 3:
            action, per3 = p3(player_state, per3)
        elif list_other[idx] == 4:
            action, per4 = p4(player_state, per4)
        else:
            raise Exception("Sai list_other.")

        env_state = stepEnv(env_state, action)
        count_turn += 1

    env_state[ENV_CHECK_END] = 1

    for p_idx in range(5):
        env_state[ENV_PHASE] = 1
        idx = int(env_state[ENV_ID_ACTION])
        player_state = getAgentState(env_state)
        if list_other[idx] == -1:
            action, per_player = p0(player_state, per_player)
        elif list_other[idx] == 1:
            action, per1 = p1(player_state, per1)
        elif list_other[idx] == 2:
            action, per2 = p2(player_state, per2)
        elif list_other[idx] == 3:
            action, per3 = p3(player_state, per3)
        elif list_other[idx] == 4:
            action, per4 = p4(player_state, per4)
        else:
            raise Exception("Sai list_other.")
        env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % NUMBER_PLAYER

    winner = False
    if np.where(list_other == -1)[0] == check_winner(env_state):
        winner = True
    else:
        winner = False

    return winner, per_player


def n_games_normal(
    p0, num_game, per_player, list_other, per1, per2, per3, per4, p1, p2, p3, p4
):
    win = 0
    for _n in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_normal(
            p0, list_other, per_player, per1, per2, per3, per4, p1, p2, p3, p4
        )
        win += winner
    return win, per_player


import importlib.util
import json, sys

import importlib.util
import json, sys

try:
    from env import SHORT_PATH
except:
    pass


def load_module_player(player):
    spec = importlib.util.spec_from_file_location(
        "Agent_player", f"{SHORT_PATH}src/Agent/{player}/Agent_player.py"
    )
    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    return module


@njit()
def check_run_under_njit(agent, perData):
    return True


def load_agent(level, *args):
    num_bot = getAgentSize() - 1

    if "_level_" not in globals():
        global _level_
        _level_ = level
        init = True
    else:
        if _level_ != level:
            _level_ = level
            init = True
        else:
            init = False

    if init:
        global _list_per_level_
        global _list_bot_level_
        _list_per_level_ = []
        _list_bot_level_ = []

        if _level_ == 0:
            _list_per_level_ = [
                np.array([[0.0]], dtype=np.float64) for _ in range(num_bot)
            ]
            _list_bot_level_ = [bot_lv0 for _ in range(num_bot)]
        else:
            env_name = sys.argv[1]
            if len(args) > 0:
                dict_level = json.load(
                    open(f"{SHORT_PATH}src/Log/check_system_about_level.json")
                )
            else:
                dict_level = json.load(open(f"{SHORT_PATH}src/Log/level_game.json"))

            if str(_level_) not in dict_level[env_name]:
                raise Exception("Hiện tại không có level này")

            lst_agent_level = dict_level[env_name][str(level)][2]

            for i in range(num_bot):
                if level == -1:
                    module_agent = load_module_player(
                        lst_agent_level[i], game_name=env_name
                    )
                    _list_per_level_.append(module_agent.DataAgent())
                else:
                    data_agent_level = np.load(
                        f"{SHORT_PATH}src/Agent/{lst_agent_level[i]}/Data/{env_name}_{level}/Train.npy",
                        allow_pickle=True,
                    )
                    module_agent = load_module_player(lst_agent_level[i])
                    _list_per_level_.append(
                        module_agent.convert_to_test(data_agent_level)
                    )
                _list_bot_level_.append(module_agent.Test)

    return _list_bot_level_, _list_per_level_


def run(
    p0: any = bot_lv0,
    num_game: int = 100,
    per_player: np.ndarray = np.array([[0.0]]),
    level: int = 0,
    *args,
):
    num_bot = getAgentSize() - 1
    list_other = np.array([-1] + [i + 1 for i in range(num_bot)])
    try:
        check_njit = check_run_under_njit(p0, per_player)
    except:
        check_njit = False

    load_agent(level, *args)

    if check_njit:
        return n_games_numba(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_per_level_[3],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
            _list_bot_level_[3],
        )
    else:
        return n_games_normal(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_per_level_[3],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
            _list_bot_level_[3],
        )
